import openmdao.api as om
from .aero_coef_comp import AeroCoefComp
from .aero_forces_comp import AeroForcesComp
from .mbi_aero_coef_comp import MBIAeroCoeffComp, setup_surrogates_all

try:
    import MBI
except ImportError:
    MBI = None


class AerodynamicsGroup(om.Group):
    """
    The purpose of the Aerodynamics is to compute the lift and
    drag forces on the aircraft.
    """
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        n = self.options['num_nodes']

        if MBI is None:
            self.add_subsystem(name='aero_coef_comp',
                               subsys=AeroCoefComp(vec_size=n, extrapolate=True,
                                                   method='lagrange3'),
                               promotes_inputs=['mach', 'alpha', 'alt', 'eta'],
                               promotes_outputs=['CL', 'CD', 'CM'])

        else:
            mbi_CL, mbi_CD, mbi_CM, mbi_num = setup_surrogates_all()
            mbi_CL.seterr(bounds='warn')
            mbi_CD.seterr(bounds='warn')
            mbi_CM.seterr(bounds='warn')

            self.add_subsystem(name='aero_coef_comp',
                               subsys=MBIAeroCoeffComp(vec_size=n, mbi_CL=mbi_CL, mbi_CD=mbi_CD,
                                                       mbi_CM=mbi_CM, mbi_num=mbi_num),
                               promotes_inputs=[('M', 'mach'), 'alpha', ('h', 'alt'), 'eta'],
                               promotes_outputs=['CL', 'CD', 'CM'])

        self.add_subsystem(name='aero_forces_comp',
                           subsys=AeroForcesComp(num_nodes=n),
                           promotes_inputs=['q', 'S', 'CL', 'CD'],
                           promotes_outputs=['L', 'D'])



import unittest

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
plt.style.use('ggplot')

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


@use_tempdirs
class TestMinTimeClimbForDocs(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_min_time_climb_for_docs_gauss_lobatto(self):
        import matplotlib.pyplot as plt

        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal

        import dymos as dm
        from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE
        from dymos.examples.plotting import plot_results

        #
        # Instantiate the problem and configure the optimization driver
        #
        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        #
        # Instantiate the trajectory and phase
        #
        traj = dm.Trajectory()

        phase = dm.Phase(ode_class=MinTimeClimbODE,
                         transcription=dm.GaussLobatto(num_segments=15, compressed=False))

        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj', traj)

        #
        # Set the options on the optimization variables
        # Note the use of explicit state units here since much of the ODE uses imperial units
        # and we prefer to solve this problem using metric units.
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                               duration_ref=100.0)

        phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6, units='m',
                        ref=1.0E3, defect_ref=1.0E3,
                        rate_source='flight_dynamics.r_dot')

        phase.add_state('h', fix_initial=True, lower=0, upper=20000.0, units='m',
                        ref=1.0E2, defect_ref=1.0E2,
                        rate_source='flight_dynamics.h_dot')

        phase.add_state('v', fix_initial=True, lower=10.0, units='m/s',
                        ref=1.0E2, defect_ref=1.0E2,
                        rate_source='flight_dynamics.v_dot')

        phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5, units='rad',
                        ref=1.0, defect_ref=1.0,
                        rate_source='flight_dynamics.gam_dot')

        phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5, units='kg',
                        ref=1.0E3, defect_ref=1.0E3,
                        rate_source='prop.m_dot')

        phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                          rate_continuity=True, rate_continuity_scaler=100.0,
                          rate2_continuity=False)

        phase.add_parameter('S', val=49.2386, units='m**2', opt=False, targets=['S'])
        phase.add_parameter('Isp', val=1600.0, units='s', opt=False, targets=['Isp'])
        phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

        #
        # Setup the boundary and path constraints
        #
        phase.add_boundary_constraint('h', loc='final', equals=20000, scaler=1.0E-3)
        phase.add_boundary_constraint('aero.mach', loc='final', equals=1.0)
        phase.add_boundary_constraint('gam', loc='final', equals=0.0)

        phase.add_path_constraint(name='h', lower=100.0, upper=20000, ref=20000)
        phase.add_path_constraint(name='aero.mach', lower=0.1, upper=1.8)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', ref=1.0)

        p.model.linear_solver = om.DirectSolver()

        #
        # Setup the problem and set the initial guess
        #
        p.setup(check=True)

        p['traj.phase0.t_initial'] = 0.0
        p['traj.phase0.t_duration'] = 500

        p.set_val('traj.phase0.states:r', phase.interp('r', [0.0, 50000.0]))
        p.set_val('traj.phase0.states:h', phase.interp('h', [100.0, 20000.0]))
        p.set_val('traj.phase0.states:v', phase.interp('v', [135.964, 283.159]))
        p.set_val('traj.phase0.states:gam', phase.interp('gam', [0.0, 0.0]))
        p.set_val('traj.phase0.states:m', phase.interp('m', [19030.468, 10000.]))
        p.set_val('traj.phase0.controls:alpha', phase.interp('alpha', [0.0, 0.0]))

        #
        # Solve for the optimal trajectory
        #
        dm.run_problem(p)

        #
        # Test the results
        #
        assert_near_equal(p.get_val('traj.phase0.t_duration'), 321.0, tolerance=1.0E-1)

        #
        # Get the explicitly simulated solution and plot the results
        #
        exp_out = traj.simulate()

        plot_results([('traj.phase0.timeseries.time', 'traj.phase0.timeseries.states:h',
                       'time (s)', 'altitude (m)'),
                      ('traj.phase0.timeseries.time', 'traj.phase0.timeseries.controls:alpha',
                       'time (s)', 'alpha (deg)')],
                     title='Supersonic Minimum Time-to-Climb Solution',
                     p_sol=p, p_sim=exp_out)

        plt.show()


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import numpy as np

from scipy.interpolate import interp1d

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal, assert_warning
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

from dymos.models.atmosphere import USatm1976Comp
from dymos.examples.min_time_climb.aero import AeroGroup
from dymos.examples.min_time_climb.prop import PropGroup
from dymos.models.eom import FlightPathEOM2D


@use_tempdirs
class TestTimeseriesOutput(unittest.TestCase):

    def test_timeseries_gl(self, test_smaller_timeseries=False):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=8, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        if test_smaller_timeseries:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665, include_timeseries=False)
        else:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        p['phase0.states:x'] = phase.interp('x', [0, 10])
        p['phase0.states:y'] = phase.interp('y', [10, 5])
        p['phase0.states:v'] = phase.interp('v', [0, 9.9])
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100])
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        gd = phase.options['transcription'].grid_data
        state_input_idxs = gd.subset_node_indices['state_input']
        control_input_idxs = gd.subset_node_indices['control_input']
        col_idxs = gd.subset_node_indices['col']

        assert_near_equal(p.get_val('phase0.time'),
                          p.get_val('phase0.timeseries.time')[:, 0])

        assert_near_equal(p.get_val('phase0.time_phase'),
                          p.get_val('phase0.timeseries.time_phase')[:, 0])

        for state in ('x', 'y', 'v'):
            assert_near_equal(p.get_val('phase0.states:{0}'.format(state)),
                              p.get_val('phase0.timeseries.states:'
                                        '{0}'.format(state))[state_input_idxs])

            assert_near_equal(p.get_val('phase0.state_interp.state_col:{0}'.format(state)),
                              p.get_val('phase0.timeseries.states:'
                                        '{0}'.format(state))[col_idxs])

        for control in ('theta',):
            assert_near_equal(p.get_val('phase0.controls:{0}'.format(control)),
                              p.get_val('phase0.timeseries.controls:'
                                        '{0}'.format(control))[control_input_idxs])

        for dp in ('g',):
            for i in range(gd.subset_num_nodes['all']):
                if test_smaller_timeseries:
                    with self.assertRaises(KeyError):
                        p.get_val('phase0.timeseries.parameters:{0}'.format(dp))
                else:
                    assert_near_equal(p.get_val('phase0.parameters:{0}'.format(dp))[0],
                                      p.get_val('phase0.timeseries.parameters:{0}'.format(dp))[i])

        # call simulate to test SolveIVP transcription
        exp_out = phase.simulate()
        if test_smaller_timeseries:
            with self.assertRaises(KeyError):
                exp_out.get_val('phase0.timeseries.parameters:{0}'.format(dp))
        else:  # no error accessing timseries.parameter
            exp_out.get_val('phase0.timeseries.parameters:{0}'.format(dp))

    def test_timeseries_gl_smaller_timeseries(self):
        self.test_timeseries_gl(test_smaller_timeseries=True)

    def test_timeseries_radau(self, test_smaller_timeseries=False):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Radau(num_segments=8, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        if test_smaller_timeseries:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665, include_timeseries=False)
        else:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.options['assembled_jac_type'] = 'csc'
        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        p['phase0.states:x'] = phase.interp('x', [0, 10])
        p['phase0.states:y'] = phase.interp('y', [10, 5])
        p['phase0.states:v'] = phase.interp('v', [0, 9.9])
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100])
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        gd = phase.options['transcription'].grid_data
        state_input_idxs = gd.subset_node_indices['state_input']
        control_input_idxs = gd.subset_node_indices['control_input']

        assert_near_equal(p.get_val('phase0.time'),
                          p.get_val('phase0.timeseries.time')[:, 0])

        assert_near_equal(p.get_val('phase0.time_phase'),
                          p.get_val('phase0.timeseries.time_phase')[:, 0])

        for state in ('x', 'y', 'v'):
            assert_near_equal(p.get_val('phase0.states:{0}'.format(state)),
                              p.get_val('phase0.timeseries.states:'
                                        '{0}'.format(state))[state_input_idxs])

        for control in ('theta',):
            assert_near_equal(p.get_val('phase0.controls:{0}'.format(control)),
                              p.get_val('phase0.timeseries.controls:'
                                        '{0}'.format(control))[control_input_idxs])

        for dp in ('g',):
            for i in range(gd.subset_num_nodes['all']):
                if test_smaller_timeseries:
                    with self.assertRaises(KeyError):
                        p.get_val('phase0.timeseries.parameters:{0}'.format(dp))
                else:
                    assert_near_equal(p.get_val('phase0.parameters:{0}'.format(dp))[0],
                                      p.get_val('phase0.timeseries.parameters:'
                                                '{0}'.format(dp))[i])

        # call simulate to test SolveIVP transcription
        exp_out = phase.simulate()
        if test_smaller_timeseries:
            with self.assertRaises(KeyError):
                exp_out.get_val('phase0.timeseries.parameters:{0}'.format(dp))
        else:  # no error accessing timseries.parameter
            exp_out.get_val('phase0.timeseries.parameters:{0}'.format(dp))

        # Test that the state rates are output in both the radau and solveivp timeseries outputs
        t_sol = p.get_val('phase0.timeseries.time')
        t_sim = exp_out.get_val('phase0.timeseries.time')

        for state_name in ('x', 'y', 'v'):
            rate_sol = p.get_val(f'phase0.timeseries.state_rates:{state_name}')
            rate_sim = exp_out.get_val(f'phase0.timeseries.state_rates:{state_name}')

            rate_t_sim = interp1d(t_sim.ravel(), rate_sim.ravel())

            assert_near_equal(rate_t_sim(t_sol), rate_sol, tolerance=1.0E-3)

    def test_timeseries_radau_smaller_timeseries(self):
        self.test_timeseries_radau(test_smaller_timeseries=True)


class MinTimeClimbODEDuplicateOutput(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_subsystem(name='atmos',
                           subsys=USatm1976Comp(num_nodes=nn),
                           promotes_inputs=['h'])

        self.add_subsystem(name='aero',
                           subsys=AeroGroup(num_nodes=nn),
                           promotes_inputs=['v', 'alpha', 'S'])

        self.connect('atmos.sos', 'aero.sos')
        self.connect('atmos.rho', 'aero.rho')

        self.add_subsystem(name='prop',
                           subsys=PropGroup(num_nodes=nn),
                           promotes_inputs=['h', 'Isp', 'throttle'])

        self.connect('aero.mach', 'prop.mach')

        self.add_subsystem(name='flight_dynamics',
                           subsys=FlightPathEOM2D(num_nodes=nn),
                           promotes_inputs=['m', 'v', 'gam', 'alpha'])

        foo = self.add_subsystem('foo', om.IndepVarComp())
        foo.add_output('rho', val=100 * np.ones(nn), units='g/cm**3')

        self.connect('aero.f_drag', 'flight_dynamics.D')
        self.connect('aero.f_lift', 'flight_dynamics.L')
        self.connect('prop.thrust', 'flight_dynamics.T')


@require_pyoptsparse(optimizer='SLSQP')
def min_time_climb(num_seg=3, transcription_class=dm.Radau, transcription_order=3,
                   force_alloc_complex=False):

    p = om.Problem(model=om.Group())

    p.driver = om.pyOptSparseDriver()

    tx = transcription_class(num_segments=num_seg, order=transcription_order)

    traj = dm.Trajectory()

    phase = dm.Phase(ode_class=MinTimeClimbODEDuplicateOutput, transcription=tx)
    traj.add_phase('phase0', phase)

    p.model.add_subsystem('traj', traj)

    phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                           duration_ref=100.0)

    phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6,
                    ref=1.0E3, defect_ref=1.0E3, units='m',
                    rate_source='flight_dynamics.r_dot')

    phase.add_state('h', fix_initial=True, lower=0, upper=20000.0,
                    ref=1.0E2, defect_ref=1.0E2, units='m',
                    rate_source='flight_dynamics.h_dot', targets=['h'])

    phase.add_state('v', fix_initial=True, lower=10.0,
                    ref=1.0E2, defect_ref=1.0E2, units='m/s',
                    rate_source='flight_dynamics.v_dot', targets=['v'])

    phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5,
                    ref=1.0, defect_ref=1.0, units='rad',
                    rate_source='flight_dynamics.gam_dot', targets=['gam'])

    phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5,
                    ref=1.0E3, defect_ref=1.0E3, units='kg',
                    rate_source='prop.m_dot', targets=['m'])

    phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                      rate_continuity=True, rate_continuity_scaler=100.0,
                      rate2_continuity=False, targets=['alpha'])

    phase.add_parameter('S', val=49.2386, units='m**2', opt=False, targets=['S'])
    phase.add_parameter('Isp', val=1600.0, units='s', opt=False, targets=['Isp'])
    phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

    phase.add_boundary_constraint('h', loc='final', equals=20000, scaler=1.0E-3)
    phase.add_boundary_constraint('aero.mach', loc='final', equals=1.0)
    phase.add_boundary_constraint('gam', loc='final', equals=0.0)

    phase.add_path_constraint(name='h', lower=100.0, upper=20000, ref=20000)
    phase.add_path_constraint(name='aero.mach', lower=0.1, upper=1.8)

    # Unnecessary but included to test capability
    phase.add_path_constraint(name='alpha', lower=-8, upper=8)
    phase.add_path_constraint(name='time', lower=0, upper=400)
    phase.add_path_constraint(name='time_phase', lower=0, upper=400)

    # Minimize time at the end of the phase
    phase.add_objective('time', loc='final', ref=1.0)

    # Add all ODE outputs to the timeseries
    phase.add_timeseries_output('*')

    p.model.linear_solver = om.DirectSolver()

    p.setup(check=True, force_alloc_complex=force_alloc_complex)

    p['traj.phase0.t_initial'] = 0.0
    p['traj.phase0.t_duration'] = 300.0

    p['traj.phase0.states:r'] = phase.interp('r', [0.0, 111319.54])
    p['traj.phase0.states:h'] = phase.interp('h', [100.0, 20000.0])
    p['traj.phase0.states:v'] = phase.interp('v', [135.964, 283.159])
    p['traj.phase0.states:gam'] = phase.interp('gam', [0.0, 0.0])
    p['traj.phase0.states:m'] = phase.interp('m', [19030.468, 16841.431])
    p['traj.phase0.controls:alpha'] = phase.interp('alpha', [0.0, 0.0])

    return p


class TestDuplicateTimeseriesGlobName(unittest.TestCase):

    def test_duplicate_timeseries_glob_name(self):
        """
        Test that the user gets a warning about multiple timeseries with the same name.
        """

        msg = "The timeseries variable name rho is duplicated in these variables: atmos.rho, " \
              "foo.rho. Disambiguate by using the add_timeseries_output output_name option."
        with assert_warning(UserWarning, msg):
            p = min_time_climb(num_seg=12, transcription_class=dm.Radau, transcription_order=3)

        with assert_warning(UserWarning, msg):
            p = min_time_climb(num_seg=12, transcription_class=dm.GaussLobatto,
                               transcription_order=3)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest
import os.path
import pathlib
import json

exclude = [
    'tests',
    'test',
    '_build',
    '.ipynb_checkpoints',
    '_srcdocs'
]

directories = []

top = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

for root, dirs, files in os.walk(top, topdown=True):
    # do not bother looking further down in excluded dirs
    dirs[:] = [d for d in dirs if d not in exclude]
    for di in dirs:
            directories.append(os.path.join(root, di))

def _get_files():

    for dir_name in directories:
        dirpath = os.path.join(top, dir_name)

        # Loop over files
        for file_name in os.listdir(dirpath):
            if not file_name.startswith('_') and file_name[-6:] == '.ipynb':
                yield dirpath + "/" + file_name

class LintJupyterOutputsTestCase(unittest.TestCase):
    """
    Check Jupyter Notebooks for outputs through execution count and recommend to remove output.
    """

    def test_output(self):
        this_file = pathlib.PurePath(__file__)
        book_dir = this_file.parent.parent

        for file in _get_files():
            print(file)
            with self.subTest(file):
                with open(file) as f:
                    json_data = json.load(f)
                    for i in json_data['cells']:
                        if 'execution_count' in i and i['execution_count'] is not None:
                            msg = "Clear output with 'jupyter nbconvert  --clear-output " \
                                  f"--inplace path_to_notebook.ipynb'"
                            self.fail(f"Output found in {file}.\n{msg}")

    def test_header(self):
        """
        Check Jupyter Notebooks for code cell installing openmdao.
        """
        header = ['try:\n',
                  '    import openmdao.api as om\n',
                  '    import dymos as dm\n',
                  'except ImportError:\n',
                  '    !python -m pip install openmdao[notebooks]\n',
                  '    !python -m pip install dymos[docs]\n',
                  '    import openmdao.api as om\n',
                  '    import dymos as dm']

        mpi_header = ['%pylab inline\n',
                      'from ipyparallel import Client, error\n',
                      'cluster=Client(profile="mpi")\n',
                      'view=cluster[:]\n',
                      'view.block=True\n',
                      '\n']
        mpi_header.extend(header)

        for file in _get_files():
            with open(file) as f:

                # This one is exempt from these lint rules.
                # if 'getting_started.ipynb'  in file:
                #     continue

                json_data = json.load(f)

                code_cells = ['code' for cell in json_data['cells'] \
                              if cell['cell_type'] == 'code']
                if len(code_cells) < 1:
                    continue

                first_block = json_data['cells'][0]['source']
                if first_block != header and first_block != mpi_header:
                    header_text = '\n'.join(header)
                    msg = f'required header not found in notebook {file}\n' \
                          f'All notebooks should contain the following block before ' \
                          f'any other code blocks:\n' \
                          f'-----------------------------------------\n' \
                          f'{header_text}\n' \
                          f'-----------------------------------------\n'
                    self.fail(msg)

                correct_tags = ['active-ipynb', 'remove-input', 'remove-output']
                msg = f"Missing metadata tags in header in notebook {file}. Found " \
                      f"headers must contain the following tags.{correct_tags}."
                try:
                    first_cell = json_data['cells'][0]['metadata']['tags']
                except KeyError:
                    msg = f"Missing metadata tags in header in notebook {file}. " \
                          f"Headers must contain the following tags: {correct_tags}."
                    self.fail(msg)

                if sorted(first_cell) != sorted(correct_tags):
                    msg = f"Incorrect header tags in notebook {file}. Found " \
                          f"{sorted(first_cell)}, should be: {sorted(correct_tags)}."
                    self.fail(msg)

    def test_assert(self):
        """
        Make sure any code cells with asserts are hidden.
        """
        for file in _get_files():
            with open(file) as f:
                json_data = json.load(f)
                blocks = json_data['cells']
                for block in blocks[1:]:

                    # Don't check markup cells
                    if block['cell_type'] != 'code':
                        continue

                    tags = block['metadata'].get('tags')
                    if tags:

                        # Don't check hidden cells
                        if 'remove-input' in tags and 'remove-output' in tags:
                            continue

                        # We allow an assert in a cell if you tag it.
                        if "allow-assert" in tags:
                            continue

                    code = ''.join(block['source'])
                    if 'assert' in code:

                        msg = f"Assert found in a code block in {file}. "
                        self.fail(msg)


if __name__ == '__main__':
    unittest.main()


import os
import unittest

import matplotlib.pyplot as plt
plt.style.use('ggplot')

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


@use_tempdirs
class TestSteadyAircraftFlightForDocs(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_steady_aircraft_for_docs(self):
        import matplotlib.pyplot as plt

        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal

        import dymos as dm

        from dymos.examples.aircraft_steady_flight.aircraft_ode import AircraftODE
        from dymos.examples.plotting import plot_results
        from dymos.utils.lgl import lgl

        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        num_seg = 15
        seg_ends, _ = lgl(num_seg + 1)

        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = traj.add_phase('phase0',
                               dm.Phase(ode_class=AircraftODE,
                                        transcription=dm.Radau(num_segments=num_seg,
                                                               segment_ends=seg_ends,
                                                               order=3, compressed=False)))

        # Pass Reference Area from an external source
        assumptions = p.model.add_subsystem('assumptions', om.IndepVarComp())
        assumptions.add_output('S', val=427.8, units='m**2')
        assumptions.add_output('mass_empty', val=1.0, units='kg')
        assumptions.add_output('mass_payload', val=1.0, units='kg')

        phase.set_time_options(fix_initial=True,
                               duration_bounds=(300, 10000),
                               duration_ref=5600)

        phase.add_state('range', units='NM',
                        rate_source='range_rate_comp.dXdt:range',
                        fix_initial=True, fix_final=False, ref=1e-3,
                        defect_ref=1e-3, lower=0, upper=2000)

        phase.add_state('mass_fuel', units='lbm',
                        rate_source='propulsion.dXdt:mass_fuel',
                        fix_initial=True, fix_final=True,
                        upper=1.5E5, lower=0.0, ref=1e2, defect_ref=1e2)

        phase.add_state('alt', units='kft',
                        rate_source='climb_rate',
                        fix_initial=True, fix_final=True,
                        lower=0.0, upper=60, ref=1e-3, defect_ref=1e-3)

        phase.add_control('climb_rate', units='ft/min', opt=True, lower=-3000, upper=3000,
                          targets=['gam_comp.climb_rate'],
                          rate_continuity=True, rate2_continuity=False)

        phase.add_control('mach', targets=['tas_comp.mach', 'aero.mach'], units=None, opt=False)

        phase.add_parameter('S',
                            targets=['aero.S', 'flight_equilibrium.S', 'propulsion.S'],
                            units='m**2')

        phase.add_parameter('mass_empty', targets=['mass_comp.mass_empty'], units='kg')
        phase.add_parameter('mass_payload', targets=['mass_comp.mass_payload'], units='kg')

        phase.add_path_constraint('propulsion.tau', lower=0.01, upper=2.0)

        p.model.connect('assumptions.S', 'traj.phase0.parameters:S')
        p.model.connect('assumptions.mass_empty', 'traj.phase0.parameters:mass_empty')
        p.model.connect('assumptions.mass_payload', 'traj.phase0.parameters:mass_payload')

        phase.add_objective('range', loc='final', ref=-1.0e-4)

        phase.add_timeseries_output('aero.CL')
        phase.add_timeseries_output('aero.CD')

        p.setup()

        p['traj.phase0.t_initial'] = 0.0
        p['traj.phase0.t_duration'] = 3600.0
        p['traj.phase0.states:range'] = phase.interp('range', ys=(0, 724.0))
        p['traj.phase0.states:mass_fuel'] = phase.interp('mass_fuel', ys=(30000, 1e-3))
        p['traj.phase0.states:alt'][:] = 10.0

        p['traj.phase0.controls:mach'][:] = 0.8

        p['assumptions.S'] = 427.8
        p['assumptions.mass_empty'] = 0.15E6
        p['assumptions.mass_payload'] = 84.02869 * 400

        dm.run_problem(p)

        assert_near_equal(p.get_val('traj.phase0.timeseries.states:range', units='NM')[-1],
                          726.85, tolerance=1.0E-2)

        exp_out = traj.simulate()

        assert_near_equal(p.get_val('traj.phase0.timeseries.CL')[0],
                          exp_out.get_val('traj.phase0.timeseries.CL')[0],
                          tolerance=1.0E-9)

        assert_near_equal(p.get_val('traj.phase0.timeseries.CD')[0],
                          exp_out.get_val('traj.phase0.timeseries.CD')[0],
                          tolerance=1.0E-9)

        plot_results([('traj.phase0.timeseries.states:range', 'traj.phase0.timeseries.states:alt',
                       'range (NM)', 'altitude (kft)'),
                      ('traj.phase0.timeseries.time', 'traj.phase0.timeseries.states:mass_fuel',
                       'time (s)', 'fuel mass (lbm)'),
                      ('traj.phase0.timeseries.time', 'traj.phase0.timeseries.CL',
                       'time (s)', 'lift coefficient'),
                      ('traj.phase0.timeseries.time', 'traj.phase0.timeseries.CD',
                       'time (s)', 'drag coefficient')],
                     title='Commercial Aircraft Optimization',
                     p_sol=p, p_sim=exp_out)

        plt.show()


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest
from openmdao.utils.testing_utils import use_tempdirs

from dymos.utils.interpolate import LagrangeBarycentricInterpolant

import numpy as np
import openmdao.api as om
import dymos as dm
from dymos.utils.lagrange import lagrange_matrices
from dymos.utils.hermite import hermite_matrices
from dymos.utils.lagrange import lagrange_matrices
from dymos.utils.lgl import lgl
from dymos.utils.lgr import lgr
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
# plt.switch_backend('Agg')  # disable plotting to the screen

from dymos.examples.oscillator.doc.oscillator_ode import OscillatorODE

NUM_SEG = 1
ORDER = 3




def plot_tangent(x, y, slope, ax, dx=1, scale=1.0, *args, **kwargs):
    x = x.ravel()
    y = y.ravel()
    slope = slope.ravel()
    x0 = x - dx * scale
    xf = x + dx * scale
    y0 = y - slope * dx * scale
    yf = y + slope * dx * scale
    ax.plot((x0, xf), (y0, yf), *args, **kwargs)


class FallEOM(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=(int,))

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('vy', shape=(nn,), units='m/s')
        self.add_output('vy_dot', shape=(nn,), units='m/s**2')
        self.add_output('y_dot', shape=(nn,), units='m/s')

        ar = np.arange(nn, dtype=int)
        self.declare_partials(of='y_dot', wrt='vy', rows=ar, cols=ar, val=1.0)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        vy = inputs['vy']
        outputs['y_dot'] = vy
        outputs['vy_dot'] = -9.80665

# Instantiate an OpenMDAO Problem instance.
prob = om.Problem()

prob.driver = om.ScipyOptimizeDriver()
prob.driver.add_recorder(om.SqliteRecorder('lgr_figures.db'))
prob.driver.recording_options['includes'] = ['*dstau*', '*rhs_col*', '*staterate*', '*timeseries*', '*timeseries2*', '*rhs_disc*']

# Instantiate a Dymos Trajectory and add it to the Problem model.
traj = dm.Trajectory()
prob.model.add_subsystem('traj', traj)

# Instantiate a Phase and add it to the Trajectory.
phase = dm.Phase(ode_class=FallEOM, transcription=dm.Radau(num_segments=NUM_SEG, order=ORDER, compressed=False))
traj.add_phase('phase0', phase)

# Tell Dymos that the duration of the phase is bounded.
phase.set_time_options(fix_initial=True, fix_duration=False)

# Tell Dymos the states to be propagated using the given ODE.
phase.add_state('y', fix_initial=True, fix_final=True, rate_source='y_dot', targets=None, units='m')
phase.add_state('vy', fix_initial=True, rate_source='vy_dot', targets=['vy'], units='m/s')

# Since we're using an optimization driver, an objective is required.  We'll minimize the final time in this case.
phase.add_objective('time', loc='final')

phase.add_timeseries('timeseries2', transcription=dm.GaussLobatto(num_segments=10, order=3))

# Setup the OpenMDAO problem
prob.setup()

# Assign values to the times and states
prob.set_val('traj.phase0.t_initial', 0.0)
prob.set_val('traj.phase0.t_duration', 2.0)

prob.set_val('traj.phase0.states:y', phase.interp('y', [100.0, 0.0]))
prob.set_val('traj.phase0.states:vy', phase.interp('vy', [0, -50]))

# Now we're using the optimization driver to iteratively run the model and vary the
# phase duration until the final y value is 0.
prob.run_driver()

state_disc_idxs = phase.options['transcription'].grid_data.subset_node_indices['state_disc']
col_idxs = phase.options['transcription'].grid_data.subset_node_indices['col']

cr = om.CaseReader('lgr_figures.db')

for i, case_name in enumerate(cr.list_cases()):

    case = cr.get_case(case_name)
    fig, axes = plt.subplots(nrows=2, ncols=1)

    axes[0].set_xlim(-.5, 5)
    axes[1].set_xlim(-.5, 5)
    axes[0].set_ylim(-50, 150)
    axes[1].set_ylim(-60, 10)

    axes[0].set_ylabel('$y$ (m)')
    axes[1].set_ylabel('$v_y$ (m/s)')
    axes[1].set_xlabel('time (s)')

    # Plot the initial state
    time = case.get_val('traj.phase0.timeseries.time')
    y = case.get_val('traj.phase0.timeseries.states:y')
    vy = case.get_val('traj.phase0.timeseries.states:vy')
    dt_dstau = case.get_val('traj.phase0.dt_dstau')

    # Plot the rates at the discretization nodes

    idxs = state_disc_idxs
    y_disc = y[state_disc_idxs]
    vy_disc = vy[state_disc_idxs]
    axes[0].plot(time[state_disc_idxs], y_disc, 'o', ms=8)
    axes[1].plot(time[state_disc_idxs], vy_disc, 'o', ms=8)

    if i == 0:
        plt.savefig('lgr_animation_0.png')

    ydot = case.get_val('traj.phase0.timeseries.state_rates:y')
    vydot = case.get_val('traj.phase0.timeseries.state_rates:vy')

    plot_tangent(time[state_disc_idxs], y[state_disc_idxs], ydot[state_disc_idxs], axes[0], color='b', scale=.25)
    plot_tangent(time[state_disc_idxs], vy[state_disc_idxs], vydot[state_disc_idxs], axes[1], color='b', scale=.25)

    if i == 0:
        plt.savefig('lgr_animation_1.png')

    # Plot the interpolating polynomials
    t_dense = np.linspace(time[0], time[-1], 100)
    L, D = lagrange_matrices(time[state_disc_idxs], t_dense)
    y_dense = L.dot(y[state_disc_idxs])
    vy_dense = L.dot(vy[state_disc_idxs])

    axes[0].plot(t_dense, y_dense, ms=None, ls=':')
    axes[1].plot(t_dense, vy_dense, ms=None, ls=':')

    if i == 0:
        plt.savefig('lgr_animation_2.png')

    # # Plot the values and rates at the collocation nodes
    # tau_s, _ = lgr(3, include_endpoint=True)
    # tau_disc = tau_s
    # tau_col = tau_s[:-1]
    # L, D = lagrange_matrices(tau_disc, tau_col)
    #
    # y_col = L.dot(y[state_disc_idxs])
    # vy_col = L.dot(vy[state_disc_idxs])
    #
    # yprime_col = D.dot(y[state_disc_idxs]) / dt_dstau[0]
    # vyprime_col = D.dot(vy[state_disc_idxs]) / dt_dstau[0]

    # # axes[0].plot(time[col_idxs], y_col, marker=(3, 0, 0), ms=8)
    # # axes[1].plot(time[col_idxs], vy_col, marker=(3, 0, 0), ms=8)
    #
    # plot_tangent(time[col_idxs], y[col_idxs], yprime_col, axes[0], color='b', scale=.25)
    # plot_tangent(time[col_idxs], vy[col_idxs], vyprime_col, axes[1], color='b', scale=.25)

    # Plot the collocation node ODE state rates
    plot_tangent(time[col_idxs], y[col_idxs], ydot[col_idxs], axes[0], color='r', scale=.25)
    plot_tangent(time[col_idxs], vy[col_idxs], vydot[col_idxs], axes[1], color='r', scale=.25)

    if i == 0:
        plt.savefig('lgr_animation_3.png')

    plt.savefig(f'lgr_solution_{i}.png')




import unittest
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal


@use_tempdirs
class TestBrachistochroneForJOSS(unittest.TestCase):

    def test_results(self):

        # Begin code for paper
        import numpy as np
        import openmdao.api as om
        import dymos as dm
        import matplotlib.pyplot as plt

        # First define a system which computes the equations of motion
        class BrachistochroneEOM(om.ExplicitComponent):
            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']

                # Inputs
                self.add_input('v', val=np.zeros(nn), units='m/s', desc='velocity')
                self.add_input('theta', val=np.zeros(nn), units='rad', desc='angle of wire')
                self.add_output('xdot', val=np.zeros(nn), units='m/s', desc='x rate of change')
                self.add_output('ydot', val=np.zeros(nn), units='m/s', desc='y rate of change')
                self.add_output('vdot', val=np.zeros(nn), units='m/s**2', desc='v rate of change')

                # Ask OpenMDAO to compute the partial derivatives using complex-step
                # with a partial coloring algorithm for improved performance
                self.declare_partials(of='*', wrt='*', method='cs')
                self.declare_coloring(wrt='*', method='cs', show_summary=True)

            def compute(self, inputs, outputs):
                v, theta = inputs.values()
                outputs['vdot'] = 9.80665 * np.cos(theta)
                outputs['xdot'] = v * np.sin(theta)
                outputs['ydot'] = -v * np.cos(theta)

        p = om.Problem()

        # Define a Trajectory object
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        # Define a Dymos Phase object with GaussLobatto Transcription
        tx = dm.GaussLobatto(num_segments=10, order=3)
        phase = dm.Phase(ode_class=BrachistochroneEOM, transcription=tx)

        traj.add_phase(name='phase0', phase=phase)

        # Set the time options
        phase.set_time_options(fix_initial=True,
                               duration_bounds=(0.5, 10.0))
        # Set the state options
        phase.add_state('x', rate_source='xdot',
                        fix_initial=True, fix_final=True)
        phase.add_state('y', rate_source='ydot',
                        fix_initial=True, fix_final=True)
        phase.add_state('v', rate_source='vdot',
                        fix_initial=True, fix_final=False)
        # Define theta as a control.
        phase.add_control(name='theta', units='rad',
                          lower=0, upper=np.pi)
        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine total
        # derivative sparsity pattern.
        # This works in conjunction with partial derivative
        # coloring to give a large speedup
        p.driver.declare_coloring()

        # Setup the problem
        p.setup()

        # Now that the OpenMDAO problem is setup, we can guess the
        # values of time, states, and controls.
        p.set_val('traj.phase0.t_duration', 2.0)

        # States and controls here use a linearly interpolated
        # initial guess along the trajectory.
        p.set_val('traj.phase0.states:x',
                  phase.interpolate(ys=[0, 10], nodes='state_input'),
                  units='m')
        p.set_val('traj.phase0.states:y',
                  phase.interpolate(ys=[10, 5], nodes='state_input'),
                  units='m')
        p.set_val('traj.phase0.states:v',
                  phase.interpolate(ys=[0, 5], nodes='state_input'),
                  units='m/s')
        # constant initial guess for control
        p.set_val('traj.phase0.controls:theta', 90, units='deg')

        # Run the driver to solve the problem and generate default plots of
        # state and control values vs time
        dm.run_problem(p, make_plots=True, simulate=True)

        # Additional custom plot of y vs x to show the actual wire shape
        fig, ax = plt.subplots(figsize=(6.4, 3.2))
        x = p.get_val('traj.phase0.timeseries.states:x', units='m')
        y = p.get_val('traj.phase0.timeseries.states:y', units='m')
        ax.plot(x,y, marker='o')
        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        fig.savefig('brachistochone_yx.png', bbox_inches='tight')
        # end code for paper
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-4)
        assert_near_equal(x[-1], 10, tolerance=1.0E-4)
        assert_near_equal(y[-1], 5, tolerance=1.0E-4)



import numpy as np
import openmdao.api as om
import dymos as dm
import matplotlib.pyplot as plt

# First define a system which computes the equations of motion
class BrachistochroneEOM(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('v', val=np.zeros(nn), units='m/s', desc='velocity')
        self.add_input('theta', val=np.zeros(nn), units='rad', desc='angle of wire')
        self.add_output('xdot', val=np.zeros(nn), units='m/s', desc='x rate of change')
        self.add_output('ydot', val=np.zeros(nn), units='m/s', desc='y rate of change')
        self.add_output('vdot', val=np.zeros(nn), units='m/s**2', desc='v rate of change')

        # Ask OpenMDAO to compute the partial derivatives using complex-step
        # with a partial coloring algorithm for improved performance
        self.declare_partials(of='*', wrt='*', method='cs')
        self.declare_coloring(wrt='*', method='cs', show_summary=True)

    def compute(self, inputs, outputs):
        v, theta = inputs.values()
        outputs['vdot'] = 9.80665 * np.cos(theta)
        outputs['xdot'] = v * np.sin(theta)
        outputs['ydot'] = -v * np.cos(theta)

p = om.Problem()

# Define a Trajectory object
traj = p.model.add_subsystem('traj', dm.Trajectory())

# Define a Dymos Phase object with GaussLobatto Transcription
tx = dm.GaussLobatto(num_segments=10, order=3)
phase = dm.Phase(ode_class=BrachistochroneEOM, transcription=tx)

traj.add_phase(name='phase0', phase=phase)

# Set the time options
phase.set_time_options(fix_initial=True,
                       duration_bounds=(0.5, 10.0))
# Set the state options
phase.add_state('x', rate_source='xdot',
                fix_initial=True, fix_final=True)
phase.add_state('y', rate_source='ydot',
                fix_initial=True, fix_final=True)
phase.add_state('v', rate_source='vdot',
                fix_initial=True, fix_final=False)
# Define theta as a control.
phase.add_control(name='theta', units='rad',
                  lower=0, upper=np.pi)
# Minimize final time.
phase.add_objective('time', loc='final')

# Set the driver.
p.driver = om.ScipyOptimizeDriver()

# Allow OpenMDAO to automatically determine total
# derivative sparsity pattern.
# This works in conjunction with partial derivative
# coloring to give a large speedup
p.driver.declare_coloring()

# Setup the problem
p.setup()

# Now that the OpenMDAO problem is setup, we can guess the
# values of time, states, and controls.
p.set_val('traj.phase0.t_duration', 2.0)

# States and controls here use a linearly interpolated
# initial guess along the trajectory.
p.set_val('traj.phase0.states:x',
          phase.interp('x', [0, 10]),
          units='m')
p.set_val('traj.phase0.states:y',
          phase.interp('y', [10, 5]),
          units='m')
p.set_val('traj.phase0.states:v',
          phase.interp('v', [0, 5]),
          units='m/s')
# constant initial guess for control
p.set_val('traj.phase0.controls:theta', 90, units='deg')

# Run the driver to solve the problem and generate default plots of
# state and control values vs time
dm.run_problem(p, make_plots=True, simulate=True)

# Additional custom plot of y vs x to show the actual wire shape
fig, ax = plt.subplots(figsize=(6.4, 3.2))
x = p.get_val('traj.phase0.timeseries.states:x', units='m')
y = p.get_val('traj.phase0.timeseries.states:y', units='m')
ax.plot(x,y, marker='o')
ax.set_xlabel('x (m)')
ax.set_ylabel('y (m)')
fig.savefig('brachistochone_yx.png', bbox_inches='tight')



__version__ = '1.3.0'

from .phase import Phase
from .transcriptions import GaussLobatto, Radau, ExplicitShooting
from .trajectory.trajectory import Trajectory
from .run_problem import run_problem
from .load_case import load_case
from .options import options



import numpy as np

import matplotlib.pyplot as plt
plt.switch_backend('Agg')
from matplotlib.patches import FancyArrowPatch, Arc

LW = 2

fig, ax = plt.subplots(1, 1, figsize=(5, 5))
ax.axis('off')

ax.set_xlim(-1, 11)
ax.set_ylim(-1, 11)

circle = plt.Circle((0, 10), radius=0.1, fc='k')
ax.add_patch(circle)
plt.text(0.2, 10.2, 'A')

circle = plt.Circle((10, 5), radius=0.1, fc='k')
ax.add_patch(circle)
plt.text(10.2, 5.2, 'B')

# Choose a to suite, compute b
a = 0.1
b = -0.5 - 10*a
c = 10

def y_wire(x):
    return a*x**2 + b*x + c, 2*a*x + b

x = np.linspace(0, 10, 100)
y, _ = y_wire(x)
plt.plot(x, y, 'b-')

# Add the bead to the wire
x = 3
y, dy_dx = y_wire(x)
plt.plot(x, y, 'ro', ms=10)

# Draw and label the gravity vector
gvec = FancyArrowPatch((x, y), (x, y-2), arrowstyle='->', mutation_scale=10, linewidth=LW, color='k')
lv_line = plt.Line2D((x, x), (y, y-2), visible=False)  # Local vertical
ax.add_patch(gvec)
plt.text(x - 0.5, y-1, 'g')

# Draw and label the velocity vector
dx = 2
dy = dy_dx * dx
vvec = FancyArrowPatch((x, y), (x+dx, y+dy), arrowstyle='->', mutation_scale=10, linewidth=LW, color='k')
ax.add_patch(vvec)
plt.text(x+dx-0.25, y+dy-0.25, 'v')

# Draw angle theta
vvec_line = plt.Line2D((x, x+dx), (y, y+dy), visible=False)
# angle_plot = get_angle_plot(lv_line, vvec_line, color='k', origin=(x, y), radius=3)
# ax.add_patch(angle_plot)
ax.text(x+0.25, y-1.25, r'$\theta$')

# Draw the axes
x = 0
y = 2
dx = 5
dy = 0
xhat = FancyArrowPatch((x, y), (x+dx, y+dy), arrowstyle='->', mutation_scale=10, linewidth=LW, color='k')
ax.add_patch(xhat)
plt.text(x+dx/2.0-0.5, y+dy/2.0-0.5, 'x')

dx = 0
dy = 5
yhat = FancyArrowPatch((x, y), (x+dx, y+dy), arrowstyle='->', mutation_scale=10, linewidth=LW, color='k')
ax.add_patch(yhat)
plt.text(x+dx/2.0-0.5, y+dy/2.0-0.5, 'y')

plt.ylim(1, 11)
plt.xlim(-0.5, 10.5)

    # plt.savefig('brachistochrone_fbd.png')

    # plt.show()



import unittest

from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal

import openmdao.api as om
import dymos as dm

from dymos.examples.racecar.combinedODE import CombinedODE
from dymos.examples.racecar.spline import get_spline, get_track_points
from dymos.examples.racecar.tracks import ovaltrack  # track curvature imports


def _run_racecar_problem(transcription, timeseries=False):
    # change track here and in curvature.py. Tracks are defined in tracks.py
    track = ovaltrack

    # generate nodes along the centerline for curvature calculation (different
    # than collocation nodes)
    points = get_track_points(track)

    # fit the centerline spline.
    finespline, gates, gatesd, curv, slope = get_spline(points, s=0.0)
    # by default 10000 points
    s_final = track.get_total_length()

    # Define the OpenMDAO problem
    p = om.Problem(model=om.Group())

    # Define a Trajectory object
    traj = dm.Trajectory()
    p.model.add_subsystem('traj', subsys=traj)

    # Define a Dymos Phase object with radau Transcription
    phase = dm.Phase(ode_class=CombinedODE,
                     transcription=transcription(num_segments=50, order=3, compressed=True))

    traj.add_phase(name='phase0', phase=phase)

    # Set the time options, in this problem we perform a change of variables. So 'time' is
    # actually 's' (distance along the track centerline)
    # This is done to fix the collocation nodes in space, which saves us the calculation of
    # the rate of change of curvature.
    # The state equations are written with respect to time, the variable change occurs in
    # timeODE.py
    phase.set_time_options(fix_initial=True, fix_duration=True, duration_val=s_final,
                           targets=['curv.s'], units='m', duration_ref=s_final,
                           duration_ref0=10)

    # Define states
    phase.add_state('t', fix_initial=True, fix_final=False, units='s', lower=0,
                    rate_source='dt_ds', ref=100)  # time
    phase.add_state('n', fix_initial=False, fix_final=False, units='m', upper=4.0, lower=-4.0,
                    rate_source='dn_ds', targets=['n'],
                    ref=4.0)  # normal distance to centerline. The bounds on n define the
    # width of the track
    phase.add_state('V', fix_initial=False, fix_final=False, units='m/s', ref=40, ref0=5,
                    rate_source='dV_ds', targets=['V'])  # velocity
    phase.add_state('alpha', fix_initial=False, fix_final=False, units='rad',
                    rate_source='dalpha_ds', targets=['alpha'],
                    ref=0.15)  # vehicle heading angle with respect to centerline
    phase.add_state('lambda', fix_initial=False, fix_final=False, units='rad',
                    rate_source='dlambda_ds', targets=['lambda'],
                    ref=0.01)  # vehicle slip angle, or angle between the axis of the vehicle
    # and velocity vector (all cars drift a little)
    phase.add_state('omega', fix_initial=False, fix_final=False, units='rad/s',
                    rate_source='domega_ds', targets=['omega'], ref=0.3)  # yaw rate
    phase.add_state('ax', fix_initial=False, fix_final=False, units='m/s**2',
                    rate_source='dax_ds', targets=['ax'], ref=8)  # longitudinal acceleration
    phase.add_state('ay', fix_initial=False, fix_final=False, units='m/s**2',
                    rate_source='day_ds', targets=['ay'], ref=8)  # lateral acceleration

    # Define Controls
    phase.add_control(name='delta', units='rad', lower=None, upper=None, fix_initial=False,
                      fix_final=False, targets=['delta'], ref=0.04)  # steering angle
    phase.add_control(name='thrust', units=None, fix_initial=False, fix_final=False, targets=[
        'thrust'])  # the thrust controls the longitudinal force of the rear tires and is
    # positive while accelerating, negative while braking

    # Performance Constraints
    pmax = 960000  # W
    phase.add_path_constraint('power', upper=pmax, ref=100000)  # engine power limit

    # The following four constraints are the tire friction limits, with 'rr' designating the
    # rear right wheel etc. This limit is computed in tireConstraintODE.py
    phase.add_path_constraint('c_rr', upper=1)
    phase.add_path_constraint('c_rl', upper=1)
    phase.add_path_constraint('c_fr', upper=1)
    phase.add_path_constraint('c_fl', upper=1)

    # Some of the vehicle design parameters are available to set here. Other parameters can
    # be found in their respective ODE files.
    phase.add_parameter('M', val=800.0, units='kg', opt=False,
                        targets=['car.M', 'tire.M', 'tireconstraint.M', 'normal.M'],
                        static_target=True)  # vehicle mass
    phase.add_parameter('beta', val=0.62, units=None, opt=False, targets=['tire.beta'],
                        static_target=True)  # brake bias
    phase.add_parameter('CoP', val=1.6, units='m', opt=False, targets=['normal.CoP'],
                        static_target=True)  # center of pressure location
    phase.add_parameter('h', val=0.3, units='m', opt=False, targets=['normal.h'],
                        static_target=True)  # center of gravity height
    phase.add_parameter('chi', val=0.5, units=None, opt=False, targets=['normal.chi'],
                        static_target=True)  # roll stiffness
    phase.add_parameter('ClA', val=4.0, units='m**2', opt=False, targets=['normal.ClA'],
                        static_target=True)  # downforce coefficient*area
    phase.add_parameter('CdA', val=2.0, units='m**2', opt=False, targets=['car.CdA'],
                        static_target=True)  # drag coefficient*area

    # Minimize final time.
    # note that we use the 'state' time instead of Dymos 'time'
    phase.add_objective('t', loc='final')

    # Add output timeseries
    if timeseries:
        phase.add_timeseries_output('*')

    # Link the states at the start and end of the phase in order to ensure a continous lap
    traj.link_phases(phases=['phase0', 'phase0'],
                     vars=['V', 'n', 'alpha', 'omega', 'lambda', 'ax', 'ay'],
                     locs=('final', 'initial'))

    # Set the driver. IPOPT or SNOPT are recommended but SLSQP might work.
    p.driver = om.pyOptSparseDriver(optimizer='IPOPT')

    p.driver.opt_settings['mu_init'] = 1e-3
    p.driver.opt_settings['max_iter'] = 500
    p.driver.opt_settings['acceptable_tol'] = 1e-3
    p.driver.opt_settings['constr_viol_tol'] = 1e-3
    p.driver.opt_settings['compl_inf_tol'] = 1e-3
    p.driver.opt_settings['acceptable_iter'] = 0
    p.driver.opt_settings['tol'] = 1e-3
    p.driver.opt_settings['nlp_scaling_method'] = 'none'
    p.driver.opt_settings['print_level'] = 5
    p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'  # for faster convergence

    # Allow OpenMDAO to automatically determine our sparsity pattern.
    # Doing so can significant speed up the execution of Dymos.
    p.driver.declare_coloring()

    # Setup the problem
    p.setup(check=True)  # force_alloc_complex=True
    # Now that the OpenMDAO problem is setup, we can set the values of the states.

    # States
    # non-zero velocity in order to protect against 1/0 errors.
    p.set_val('traj.phase0.states:V', phase.interp('V', [20, 20]), units='m/s')
    p.set_val('traj.phase0.states:lambda', phase.interp('lambda', [0.0, 0.0]), units='rad')
    # all other states start at 0
    p.set_val('traj.phase0.states:omega', phase.interp('omega', [0.0, 0.0]), units='rad/s')
    p.set_val('traj.phase0.states:alpha', phase.interp('alpha', [0.0, 0.0]), units='rad')
    p.set_val('traj.phase0.states:ax', phase.interp('ax', [0.0, 0.0]), units='m/s**2')
    p.set_val('traj.phase0.states:ay', phase.interp('ay', [0.0, 0.0]), units='m/s**2')
    p.set_val('traj.phase0.states:n', phase.interp('n', [0.0, 0.0]), units='m')
    # initial guess for what the final time should be
    p.set_val('traj.phase0.states:t', phase.interp('t', [0.0, 100.0]), units='s')

    # Controls
    p.set_val('traj.phase0.controls:delta', phase.interp('delta', [0.0, 0.0]), units='rad')
    p.set_val('traj.phase0.controls:thrust', phase.interp('thrust', [0.1, 0.1]), units=None)
    # a small amount of thrust can speed up convergence

    dm.run_problem(p, run_driver=True, simulate=False, make_plots=False)
    print('Optimization finished')

    t = p.get_val('traj.phase0.timeseries.states:t')
    assert_near_equal(t[-1], 22.2657, tolerance=0.01)


@use_tempdirs
class BenchmarkRacecar(unittest.TestCase):
    """ Benchmarks for various permutations of the racecar problem."""

    def benchmark_gausslobatto_notimeseries(self):
        _run_racecar_problem(dm.GaussLobatto, timeseries=False)

    def benchmark_gausslobatto_timeseries(self):
        _run_racecar_problem(dm.GaussLobatto, timeseries=True)

    def benchmark_radau_notimeseries(self):
        _run_racecar_problem(dm.Radau, timeseries=False)

    def benchmark_radau_timeseries(self):
        _run_racecar_problem(dm.Radau, timeseries=True)



import numpy as np
from scipy.linalg import block_diag

import openmdao.api as om
from openmdao.utils.units import unit_conversion

from ...transcriptions.grid_data import GridData
from ...utils.lagrange import lagrange_matrices
from ...options import options as dymos_options


class TimeseriesOutputCompBase(om.ExplicitComponent):
    """
    Class definition of the TimeseriesOutputCompBase.

    TimeseriesOutputComp collects variable values from the phase and provides them in chronological
    order as outputs.  Some phase types don't internally have access to a contiguous array of all
    values of a given variable in the phase.  For instance, the GaussLobatto pseudospectral has
    separate arrays of variable values at discretization and collocation nodes.  These values
    need to be interleaved to provide a time series.  Pseudospectral techniques provide timeseries
    data at 'all' nodes, while ExplicitPhase provides values at the step boundaries.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self._timeseries_outputs = []

        self._vars = {}

        self.options.declare('input_grid_data',
                             types=GridData,
                             desc='Container object for grid on which inputs are provided.')

        self.options.declare('output_grid_data',
                             types=GridData,
                             allow_none=True,
                             default=None,
                             desc='Container object for grid on which outputs are interpolated.')

        self.options.declare('output_subset',
                             types=str,
                             default='all',
                             desc='Name of the node subset at which outputs are desired.')


class PseudospectralTimeseriesOutputComp(TimeseriesOutputCompBase):
    """
    Class definition of the PseudospectralTimeseriesOutputComp.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super(PseudospectralTimeseriesOutputComp, self).__init__(**kwargs)

        self.input_num_nodes = 0
        self.output_num_nodes = 0

        # Sources is used internally to map the source of a connection to the timeseries to
        # the corresponding input variable.  This is used to ensure that we don't need to connect
        # the same source to this timeseries multiple times.
        self._sources = {}

        # Used to track conversion factors for instances when one output that relies on an input
        # from another variable has potentially different units
        self._units = {}
        self._conversion_factors = {}

    def setup(self):
        """
        Define the independent variables as output variables.
        """
        igd = self.options['input_grid_data']
        ogd = self.options['output_grid_data']
        output_subset = self.options['output_subset']

        if ogd is None:
            ogd = igd

        self.input_num_nodes = igd.num_nodes
        self.output_num_nodes = ogd.subset_num_nodes[output_subset]

        # Build the interpolation matrix which maps from the input grid to the output grid.
        # Rather than a single phase-wide interpolating polynomial, map each segment.
        # To do this, find the nodes in the output grid which fall in each segment of the input
        # grid.  Then build a Lagrange interpolating polynomial for that segment
        L_blocks = []
        output_nodes_ptau = ogd.node_ptau[ogd.subset_node_indices[output_subset]].tolist()

        for iseg in range(igd.num_segments):
            i1, i2 = igd.segment_indices[iseg]
            iptau_segi = igd.node_ptau[i1:i2]
            istau_segi = igd.node_stau[i1:i2]

            # The indices of the output grid that fall within this segment of the input grid
            if ogd is igd:
                optau_segi = iptau_segi
            else:
                ptau_hi = igd.segment_ends[iseg+1]
                if iseg < igd.num_segments - 1:
                    idxs_in_iseg = np.where(output_nodes_ptau <= ptau_hi)[0]
                else:
                    idxs_in_iseg = np.arange(len(output_nodes_ptau))
                optau_segi = np.asarray(output_nodes_ptau)[idxs_in_iseg]
                # Remove the captured nodes so we don't accidentally include them again
                output_nodes_ptau = output_nodes_ptau[len(idxs_in_iseg):]

            # # Now get the output nodes which fall in iseg in iseg's segment tau space.
            ostau_segi = 2.0 * (optau_segi - iptau_segi[0]) / (iptau_segi[-1] - iptau_segi[0]) - 1

            # Create the interpolation matrix and add it to the blocks
            L, _ = lagrange_matrices(istau_segi, ostau_segi)
            L_blocks.append(L)

        self.interpolation_matrix = block_diag(*L_blocks)

        for (name, kwargs) in self._timeseries_outputs:
            units = kwargs['units']
            desc = kwargs['units']
            shape = kwargs['shape']
            self._add_output_configure(name, units, shape, desc)

    def _add_output_configure(self, name, units, shape, desc='', src=None):
        """
        Add a single timeseries output.

        Can be called by parent groups in configure.

        Parameters
        ----------
        name : str
            name of the variable in this component's namespace.
        shape : int or tuple or list or None
            Shape of this variable, only required if val is not an array.
            Default is None.
        units : str or None
            Units in which the output variables will be provided to the component during execution.
            Default is None, which means it has no units.
        desc : str
            description of the timeseries output variable.
        src : str
            The src path of the variables input, used to prevent redundant inputs.

        Returns
        -------
        bool
            True if a new input was added for the output, or False if it reuses an existing input.
        """
        input_num_nodes = self.input_num_nodes
        output_num_nodes = self.output_num_nodes
        added_source = False

        if name in self._vars:
            return False

        if src in self._sources:
            # If we're already pulling the source into this timeseries, use that as the
            # input for this output.
            input_name = self._sources[src]
            input_units = self._units[input_name]
        else:
            input_name = f'input_values:{name}'
            self.add_input(input_name,
                           shape=(input_num_nodes,) + shape,
                           units=units, desc=desc)
            self._sources[src] = input_name
            input_units = self._units[input_name] = units
            added_source = True

        output_name = name
        self.add_output(output_name,
                        shape=(output_num_nodes,) + shape,
                        units=units, desc=desc)

        self._vars[name] = (input_name, output_name, shape)

        size = np.prod(shape)
        val_jac = np.zeros((output_num_nodes, size, input_num_nodes, size))

        for i in range(size):
            val_jac[:, i, :, i] = self.interpolation_matrix

        val_jac = val_jac.reshape((output_num_nodes * size, input_num_nodes * size),
                                  order='C')

        val_jac_rows, val_jac_cols = np.where(val_jac != 0)

        rs, cs = val_jac_rows, val_jac_cols

        # There's a chance that the input for this output was pulled from another variable with
        # different units, so account for that with a conversion.
        if None in {input_units, units}:
            scale = 1.0
            offset = 0
        else:
            scale, offset = unit_conversion(input_units, units)
        self._conversion_factors[output_name] = scale, offset

        self.declare_partials(of=output_name,
                              wrt=input_name,
                              rows=rs, cols=cs, val=scale * val_jac[rs, cs])

        return added_source

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        for (input_name, output_name, _) in self._vars.values():
            scale, offset = self._conversion_factors[output_name]

            inp = inputs[input_name]
            if len(inp.shape) > 2:
                # Dot product always performs the sum product over axis 2.
                inp = inp.swapaxes(0, 1)

            interp_vals = self.interpolation_matrix.dot(inp)
            outputs[output_name] = scale * (interp_vals + offset)



import openmdao.api as om
import dymos as dm

from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


prob = om.Problem()

prob.driver = om.pyOptSparseDriver(optimizer='SLSQP')

tx = dm.transcriptions.ExplicitShooting(num_segments=5, grid='gauss-lobatto', method='rk4',
                                        order=3, num_steps_per_segment=10, compressed=True)

phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

# automatically discover states
phase.set_state_options('x', fix_initial=True)
phase.set_state_options('y', fix_initial=True)
phase.set_state_options('v', fix_initial=True)

phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1, upper=9.80665)
phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9)

phase.add_boundary_constraint('x', loc='final', equals=10.0)
phase.add_boundary_constraint('y', loc='final', equals=5.0)

phase.add_timeseries_output('*')

prob.model.add_subsystem('phase0', phase)

phase.add_objective('time', loc='final')

prob.setup(force_alloc_complex=True)

prob.set_val('phase0.t_initial', 0.0)
prob.set_val('phase0.t_duration', 2)
prob.set_val('phase0.states:x', 0.0)
prob.set_val('phase0.states:y', 10.0)
prob.set_val('phase0.states:v', 1.0E-6)
prob.set_val('phase0.parameters:g', 1.0, units='m/s**2')
prob.set_val('phase0.controls:theta', phase.interp('theta', ys=[0.01, 90]), units='deg')

dm.run_problem(prob, run_driver=True)



import numpy as np
from dymos.transcriptions.common.timeseries_output_comp import TimeseriesOutputCompBase


class ExplicitTimeseriesOutputComp(TimeseriesOutputCompBase):
    """
    Class definition for SolveIVPTimeseriesOutputComp.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """

    def setup(self):
        """
        Define the independent variables as output variables.
        """
        grid_data = self.options['input_grid_data']
        self.num_nodes = 2 * grid_data.num_segments

    def _add_output_configure(self, name, units, shape, desc):
        """
        Add a single timeseries output.

        Can be called by parent groups in configure.

        Parameters
        ----------
        name : str
            name of the variable in this component's namespace.
        shape : int or tuple or list or None
            Shape of this variable, only required if val is not an array.
            Default is None.
        units : str or None
            Units in which the output variables will be provided to the component during execution.
            Default is None, which means it has no units.
        desc : str
            description of the timeseries output variable.
        """
        num_nodes = self.num_nodes

        input_name = f'all_values:{name}'
        self.add_input(input_name,
                       shape=(num_nodes,) + shape,
                       units=units, desc=desc)

        output_name = name
        self.add_output(output_name,
                        shape=(num_nodes,) + shape,
                        units=units, desc=desc)

        self._vars[name] = (input_name, output_name, shape)

        size = np.prod(shape, dtype=int)
        ar = np.arange(size * num_nodes, dtype=int)
        self.declare_partials(of=output_name, wrt=input_name, rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        for (input_name, output_name, _) in self._vars.values():
            outputs[output_name] = inputs[input_name]



import unittest

import matplotlib.pyplot as plt
plt.switch_backend('Agg')

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


@use_tempdirs
class TestTwoPhaseCannonballForDocs(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_two_phase_cannonball_for_docs(self):
        import numpy as np
        from scipy.interpolate import interp1d

        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal

        import dymos as dm
        from dymos.models.atmosphere.atmos_1976 import USatm1976Data

        #############################################
        # Component for the design part of the model
        #############################################
        class CannonballSizeComp(om.ExplicitComponent):
            """
            Compute the area and mass of a cannonball with a given radius and density.

            Notes
            -----
            This component is not vectorized with 'num_nodes' as is the usual way
            with Dymos, but is instead intended to compute a scalar mass and reference
            area from scalar radius and density inputs. This component does not reside
            in the ODE but instead its outputs are connected to the trajectory via
            input design parameters.
            """
            def setup(self):
                self.add_input(name='radius', val=1.0, desc='cannonball radius', units='m')
                self.add_input(name='dens', val=7870., desc='cannonball density', units='kg/m**3')

                self.add_output(name='mass', shape=(1,), desc='cannonball mass', units='kg')
                self.add_output(name='S', shape=(1,), desc='aerodynamic reference area', units='m**2')

                self.declare_partials(of='mass', wrt='dens')
                self.declare_partials(of='mass', wrt='radius')

                self.declare_partials(of='S', wrt='radius')

            def compute(self, inputs, outputs):
                radius = inputs['radius']
                dens = inputs['dens']

                outputs['mass'] = (4/3.) * dens * np.pi * radius ** 3
                outputs['S'] = np.pi * radius ** 2

            def compute_partials(self, inputs, partials):
                radius = inputs['radius']
                dens = inputs['dens']

                partials['mass', 'dens'] = (4/3.) * np.pi * radius ** 3
                partials['mass', 'radius'] = 4. * dens * np.pi * radius ** 2

                partials['S', 'radius'] = 2 * np.pi * radius

        #############################################
        # Build the ODE class
        #############################################
        class CannonballODE(om.ExplicitComponent):
            """
            Cannonball ODE assuming flat earth and accounting for air resistance
            """

            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']

                # static parameters
                self.add_input('m', units='kg')
                self.add_input('S', units='m**2')
                # 0.5 good assumption for a sphere
                self.add_input('CD', 0.5)

                # time varying inputs
                self.add_input('h', units='m', shape=nn)
                self.add_input('v', units='m/s', shape=nn)
                self.add_input('gam', units='rad', shape=nn)

                # state rates
                self.add_output('v_dot', shape=nn, units='m/s**2', tags=['dymos.state_rate_source:v'])
                self.add_output('gam_dot', shape=nn, units='rad/s', tags=['dymos.state_rate_source:gam'])
                self.add_output('h_dot', shape=nn, units='m/s', tags=['dymos.state_rate_source:h'])
                self.add_output('r_dot', shape=nn, units='m/s', tags=['dymos.state_rate_source:r'])
                self.add_output('ke', shape=nn, units='J')

                # Ask OpenMDAO to compute the partial derivatives using complex-step
                # with a partial coloring algorithm for improved performance, and use
                # a graph coloring algorithm to automatically detect the sparsity pattern.
                self.declare_coloring(wrt='*', method='cs')

                alt_data = USatm1976Data.alt * om.unit_conversion('ft', 'm')[0]
                rho_data = USatm1976Data.rho * om.unit_conversion('slug/ft**3', 'kg/m**3')[0]
                self.rho_interp = interp1d(np.array(alt_data, dtype=complex),
                                           np.array(rho_data, dtype=complex),
                                           kind='linear')

            def compute(self, inputs, outputs):

                gam = inputs['gam']
                v = inputs['v']
                h = inputs['h']
                m = inputs['m']
                S = inputs['S']
                CD = inputs['CD']

                GRAVITY = 9.80665  # m/s**2

                # handle complex-step gracefully from the interpolant
                if np.iscomplexobj(h):
                    rho = self.rho_interp(inputs['h'])
                else:
                    rho = self.rho_interp(inputs['h']).real

                q = 0.5*rho*inputs['v']**2
                qS = q * S
                D = qS * CD
                cgam = np.cos(gam)
                sgam = np.sin(gam)
                outputs['v_dot'] = - D/m-GRAVITY*sgam
                outputs['gam_dot'] = -(GRAVITY/v)*cgam
                outputs['h_dot'] = v*sgam
                outputs['r_dot'] = v*cgam
                outputs['ke'] = 0.5*m*v**2

        #############################################
        # Setup the Dymos problem
        #############################################

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        p.model.add_subsystem('size_comp', CannonballSizeComp(),
                              promotes_inputs=['radius', 'dens'])
        p.model.set_input_defaults('dens', val=7.87, units='g/cm**3')
        p.model.add_design_var('radius', lower=0.01, upper=0.10,
                               ref0=0.01, ref=0.10, units='m')

        traj = p.model.add_subsystem('traj', dm.Trajectory())

        transcription = dm.Radau(num_segments=5, order=3, compressed=True)
        ascent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        ascent = traj.add_phase('ascent', ascent)

        # All initial states except flight path angle are fixed
        # Final flight path angle is fixed (we will set it to zero
        # so that the phase ends at apogee).
        # The output of the ODE which provides the rate source for each state
        # is obtained from the tags used on those outputs in the ODE.
        # The units of the states are automatically inferred by multiplying the units
        # of those rates by the time units.
        ascent.set_time_options(fix_initial=True, duration_bounds=(1, 100),
                                duration_ref=100, units='s')
        ascent.set_state_options('r', fix_initial=True, fix_final=False)
        ascent.set_state_options('h', fix_initial=True, fix_final=False)
        ascent.set_state_options('gam', fix_initial=False, fix_final=True)
        ascent.set_state_options('v', fix_initial=False, fix_final=False)

        ascent.add_parameter('S', units='m**2', static_target=True)
        ascent.add_parameter('m', units='kg', static_target=True)

        # Limit the muzzle energy
        ascent.add_boundary_constraint('ke', loc='initial',
                                       upper=400000, lower=0, ref=100000)

        # Second Phase (descent)
        transcription = dm.GaussLobatto(num_segments=5, order=3, compressed=True)
        descent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        traj.add_phase('descent', descent)

        # All initial states and time are free, since
        #    they will be linked to the final states of ascent.
        # Final altitude is fixed, because we will set
        #    it to zero so that the phase ends at ground impact)
        descent.set_time_options(initial_bounds=(.5, 100), duration_bounds=(.5, 100),
                                 duration_ref=100, units='s')
        descent.add_state('r')
        descent.add_state('h', fix_initial=False, fix_final=True)
        descent.add_state('gam', fix_initial=False, fix_final=False)
        descent.add_state('v', fix_initial=False, fix_final=False)

        descent.add_parameter('S', units='m**2', static_target=True)
        descent.add_parameter('m', units='kg', static_target=True)

        descent.add_objective('r', loc='final', scaler=-1.0)

        # Add internally-managed design parameters to the trajectory.
        traj.add_parameter('CD',
                           targets={'ascent': ['CD'], 'descent': ['CD']},
                           val=0.5, units=None, opt=False, static_target=True)

        # Add externally-provided design parameters to the trajectory.
        # In this case, we connect 'm' to pre-existing input parameters
        # named 'mass' in each phase.
        traj.add_parameter('m', units='kg', val=1.0,
                           targets={'ascent': 'mass', 'descent': 'mass'}, static_target=True)

        # In this case, by omitting targets, we're connecting these
        # parameters to parameters with the same name in each phase.
        traj.add_parameter('S', units='m**2', val=0.005, static_target=True)

        # Link Phases (link time and all state variables)
        traj.link_phases(phases=['ascent', 'descent'], vars=['*'])

        # Issue Connections
        p.model.connect('size_comp.mass', 'traj.parameters:m')
        p.model.connect('size_comp.S', 'traj.parameters:S')

        # A linear solver at the top level can improve performance.
        p.model.linear_solver = om.DirectSolver()

        # Finish Problem Setup
        p.setup()

        #############################################
        # Set constants and initial guesses
        #############################################
        p.set_val('radius', 0.05, units='m')
        p.set_val('dens', 7.87, units='g/cm**3')

        p.set_val('traj.parameters:CD', 0.5)

        p.set_val('traj.ascent.t_initial', 0.0)
        p.set_val('traj.ascent.t_duration', 10.0)

        p.set_val('traj.ascent.states:r', ascent.interp('r', [0, 100]))
        p.set_val('traj.ascent.states:h', ascent.interp('h', [0, 100]))
        p.set_val('traj.ascent.states:v', ascent.interp('v', [200, 150]))
        p.set_val('traj.ascent.states:gam', ascent.interp('gam', [25, 0]), units='deg')

        p.set_val('traj.descent.t_initial', 10.0)
        p.set_val('traj.descent.t_duration', 10.0)

        p.set_val('traj.descent.states:r', descent.interp('r', [100, 200]))
        p.set_val('traj.descent.states:h', descent.interp('h', [100, 0]))
        p.set_val('traj.descent.states:v', descent.interp('v', [150, 200]))
        p.set_val('traj.descent.states:gam', descent.interp('gam', [0, -45]), units='deg')

        #####################################################
        # Run the optimization and final explicit simulation
        #####################################################
        dm.run_problem(p)

        assert_near_equal(p.get_val('traj.descent.states:r')[-1],
                          3183.25, tolerance=1.0E-2)

        # use the explicit simulation to check the final collocation solution accuracy
        exp_out = traj.simulate()

        #############################################
        # Plot the results
        #############################################
        rad = p.get_val('radius', units='m')[0]
        print(f'optimal radius: {rad} m ')
        mass = p.get_val('size_comp.mass', units='kg')[0]
        print(f'cannonball mass: {mass} kg ')
        angle = p.get_val('traj.ascent.timeseries.states:gam', units='deg')[0, 0]
        print(f'launch angle: {angle} deg')
        max_range = p.get_val('traj.descent.timeseries.states:r')[-1, 0]
        print(f'maximum range: {max_range} m')

        fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(10, 6))

        time_imp = {'ascent': p.get_val('traj.ascent.timeseries.time'),
                    'descent': p.get_val('traj.descent.timeseries.time')}

        time_exp = {'ascent': exp_out.get_val('traj.ascent.timeseries.time'),
                    'descent': exp_out.get_val('traj.descent.timeseries.time')}

        r_imp = {'ascent': p.get_val('traj.ascent.timeseries.states:r'),
                 'descent': p.get_val('traj.descent.timeseries.states:r')}

        r_exp = {'ascent': exp_out.get_val('traj.ascent.timeseries.states:r'),
                 'descent': exp_out.get_val('traj.descent.timeseries.states:r')}

        h_imp = {'ascent': p.get_val('traj.ascent.timeseries.states:h'),
                 'descent': p.get_val('traj.descent.timeseries.states:h')}

        h_exp = {'ascent': exp_out.get_val('traj.ascent.timeseries.states:h'),
                 'descent': exp_out.get_val('traj.descent.timeseries.states:h')}

        axes.plot(r_imp['ascent'], h_imp['ascent'], 'bo')

        axes.plot(r_imp['descent'], h_imp['descent'], 'ro')

        axes.plot(r_exp['ascent'], h_exp['ascent'], 'b--')

        axes.plot(r_exp['descent'], h_exp['descent'], 'r--')

        axes.set_xlabel('range (m)')
        axes.set_ylabel('altitude (m)')

        fig, axes = plt.subplots(nrows=4, ncols=1, figsize=(10, 6))
        states = ['r', 'h', 'v', 'gam']
        for i, state in enumerate(states):
            x_imp = {'ascent': p.get_val(f'traj.ascent.timeseries.states:{state}'),
                     'descent': p.get_val(f'traj.descent.timeseries.states:{state}')}

            x_exp = {'ascent': exp_out.get_val(f'traj.ascent.timeseries.states:{state}'),
                     'descent': exp_out.get_val(f'traj.descent.timeseries.states:{state}')}

            axes[i].set_ylabel(state)

            axes[i].plot(time_imp['ascent'], x_imp['ascent'], 'bo')
            axes[i].plot(time_imp['descent'], x_imp['descent'], 'ro')
            axes[i].plot(time_exp['ascent'], x_exp['ascent'], 'b--')
            axes[i].plot(time_exp['descent'], x_exp['descent'], 'r--')

        params = ['m', 'S']
        fig, axes = plt.subplots(nrows=6, ncols=1, figsize=(12, 6))
        for i, param in enumerate(params):
            p_imp = {
                'ascent': p.get_val(f'traj.ascent.timeseries.parameters:{param}'),
                'descent': p.get_val(f'traj.descent.timeseries.parameters:{param}')}

            p_exp = {'ascent': exp_out.get_val(f'traj.ascent.timeseries.parameters:{param}'),
                     'descent': exp_out.get_val(f'traj.descent.timeseries.parameters:{param}')}

            axes[i].set_ylabel(param)

            axes[i].plot(time_imp['ascent'], p_imp['ascent'], 'bo')
            axes[i].plot(time_imp['descent'], p_imp['descent'], 'ro')
            axes[i].plot(time_exp['ascent'], p_exp['ascent'], 'b--')
            axes[i].plot(time_exp['descent'], p_exp['descent'], 'r--')

        plt.show()


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest
from openmdao.utils.testing_utils import use_tempdirs

from dymos.utils.interpolate import LagrangeBarycentricInterpolant

import numpy as np
import openmdao.api as om
import dymos as dm
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
# plt.switch_backend('Agg')  # disable plotting to the screen

from dymos.examples.oscillator.doc.oscillator_ode import OscillatorODE

NUM_SEG = 1
ORDER = 5
NUM_FRAMES = 20
ORDER = 3

# Instantiate an OpenMDAO Problem instance.
prob = om.Problem()

# We need an optimization driver.  To solve this simple problem ScipyOptimizerDriver will work.
prob.driver = om.ScipyOptimizeDriver()

# Instantiate a Dymos Trajectory and add it to the Problem model.
traj = dm.Trajectory()
prob.model.add_subsystem('traj', traj)

# Instantiate a Phase and add it to the Trajectory.
phase = dm.Phase(ode_class=OscillatorODE, transcription=dm.Radau(num_segments=NUM_SEG, order=ORDER, compressed=False))
traj.add_phase('phase0', phase)

# Tell Dymos that the duration of the phase is bounded.
phase.set_time_options(fix_initial=True, fix_duration=True)

# Tell Dymos the states to be propagated using the given ODE.
phase.add_state('x', fix_initial=True, rate_source='v', targets=['x'], units='m')
phase.add_state('v', fix_initial=True, rate_source='v_dot', targets=['v'], units='m/s')

# The spring constant, damping coefficient, and mass are inputs to the system that are constant throughout the phase.
phase.add_parameter('k', units='N/m', targets=['k'])
phase.add_parameter('c', units='N*s/m', targets=['c'])
phase.add_parameter('m', units='kg', targets=['m'])

# secondary "dense" timeseries
phase.add_timeseries('timeseries2', transcription=dm.Radau(num_segments=NUM_SEG, order=41, compressed=False))

# Since we're using an optimization driver, an objective is required.  We'll minimize the final time in this case.
phase.add_objective('time', loc='final')

# Setup the OpenMDAO problem
prob.setup()

# Assign values to the times and states
prob.set_val('traj.phase0.t_initial', 0.0)
prob.set_val('traj.phase0.t_duration', 5.0)

prob.set_val('traj.phase0.states:x', 10.0)
prob.set_val('traj.phase0.states:v', 0.0)

prob.set_val('traj.phase0.parameters:k', 1.0)
prob.set_val('traj.phase0.parameters:c', 0.5)
prob.set_val('traj.phase0.parameters:m', 1.0)

starts= {}

starts['x'] = prob.get_val('traj.phase0.states:x')
starts['v'] = prob.get_val('traj.phase0.states:v')

# Now we're using the optimization driver to iteratively run the model and vary the
# phase duration until the final y value is 0.
prob.run_driver()

# Perform an explicit simulation of our ODE from the initial conditions.
sim_out = traj.simulate(times_per_seg=50)

# Plot the state values obtained from the phase timeseries objects in the simulation output.
t_sol = prob.get_val('traj.phase0.timeseries.time')
t_sim = sim_out.get_val('traj.phase0.timeseries.time')
t_dense = prob.get_val('traj.phase0.timeseries2.time')

all_idxs = phase.options['transcription'].grid_data.subset_segment_indices['all']

states = ['x', 'v']
solutions = {}
histories = {}

for i, state in enumerate(states):
    state_sol = prob.get_val(f'traj.phase0.timeseries.states:{state}')
    state_sim = sim_out.get_val(f'traj.phase0.timeseries.states:{state}')
    state_dense = prob.get_val(f'traj.phase0.timeseries2.states:{state}')
    # sol = axes[i].plot(t_sol, state_sol, 'o')
    # sim = axes[i].plot(t_sim, state_sim, '-')
    # dense = axes[i].plot(t_dense, state_dense, '--', color='#CCCCCC')
    # axes[i].set_ylabel(state)
    solutions[state] = state_sol
    histories[state] = np.linspace(starts[state], solutions[state], NUM_FRAMES)

fig, axes = plt.subplots(len(states), 1)

for i in range(NUM_FRAMES):
    for j, state in enumerate(states):
        prob.set_val(f'traj.phase0.states:{state}', histories[state][i, ...])

    prob.run_model()

    for j, state in enumerate(states):
        state_sol = prob.get_val(f'traj.phase0.timeseries.states:{state}')
        state_dense = prob.get_val(f'traj.phase0.timeseries2.states:{state}')
        sol = axes[j].plot(t_sol, state_sol, 'o')
        dense = axes[j].plot(t_dense, state_dense, '--', color='#CCCCCC')

t_data = []
x_data = []
v_data = []

fig, axes = plt.subplots(len(states), 1)

x_sol_line, = axes[0].plot(t_data, x_data, 'o')
v_sol_line, = axes[1].plot(t_data, v_data, 'o')
# x_dense_line, = axes[0].plot(t_data, x_data, '--')
# v_dense_line, = axes[1].plot(t_data, v_data, '--')
lines = x_sol_line, v_sol_line

def init():
    return lines

def update(frame):
    # xdata.append(frame)
    # ydata.append(np.sin(frame))

    x_sol_line, x_dense_line, v_sol_line, v_dense_line = lines

    for j, state in enumerate(states):
        prob.set_val(f'traj.phase0.states:{state}', histories[state][frame, ...])

    prob.run_model()

    x_sol = prob.get_val('traj.phase0.timeseries.states:x')
    # x_dense = prob.get_val('traj.phase0.timeseries2.states:x')

    v_sol = prob.get_val('traj.phase0.timeseries.states:v')
    # v_dense = prob.get_val('traj.phase0.timeseries2.states:v')

    x_sol_line.set_ydata(x_sol)
    # x_dense_line.set_ydata(x_dense)
    v_sol_line.set_ydata(v_sol)
    # v_dense_line.set_ydata(v_dense)

    return x_sol_line, v_sol_line
    # return

ani = FuncAnimation(fig, update, frames=range(NUM_FRAMES),
                    init_func=init, blit=True)

plt.show()

    # for j in range(NUM_SEG):
    #     start_idx, end_idx = all_idxs[j, :]
    #     t_seg_start = t_sol.ravel()[start_idx]
    #     t_seg_end = t_sol.ravel()[end_idx - 1]
        # print(t_seg_start, t_seg_end)
        # print(state_sol[start_idx:end_idx])
        # print(t_seg_start, t_seg_end)
        # nodes_stau = phase.options['transcription'].grid_data.node_stau[start_idx:end_idx]
        # lbi = LagrangeBarycentricInterpolant(nodes=nodes_stau, shape=(1,))
        # # print(nodes_stau)
        # lbi.setup(t_seg_start, t_seg_end, state_sol[start_idx:end_idx])
        # t_interp = np.linspace(t_seg_start, t_seg_end, 100)
        # with np.printoptions(linewidth=1024, edgeitems=500):
        #     sol_interp = lbi.eval(t_interp)
        #     soldot_interp = lbi.eval_deriv(0.0)
        # axes[i].plot(t_interp, sol_interp, 'k-')



# axes[-1].set_xlabel('time (s)')
# # fig.legend((sol[0], sim[0]), ('solution', 'simulation'), 'lower right', ncol=2)
# plt.tight_layout()
# plt.show()
fig, axes = plt.subplots(len(states), 1)
for i, state in enumerate(states):
    state_sol = prob.get_val(f'traj.phase0.timeseries.states:{state}')
    state_sim = sim_out.get_val(f'traj.phase0.timeseries.states:{state}')
    sol = axes[i].plot(t_sol, state_sol, 'o')
    # sim = axes[i].plot(t_sim, state_sim, '-')
    axes[i].set_ylabel(state)

    for j in range(NUM_SEG):
        start_idx, end_idx = all_idxs[j, :]
        t_seg_start = t_sol.ravel()[start_idx]
        t_seg_end = t_sol.ravel()[end_idx - 1]
        # print(t_seg_start, t_seg_end)
        # print(state_sol[start_idx:end_idx])
        # print(t_seg_start, t_seg_end)
        nodes_stau = phase.options['transcription'].grid_data.node_stau[start_idx:end_idx]
        lbi = LagrangeBarycentricInterpolant(nodes=nodes_stau, shape=(1,))
        # print(nodes_stau)
        lbi.setup(t_seg_start, t_seg_end, state_sol[start_idx:end_idx])
        t_interp = np.linspace(t_seg_start, t_seg_end, 100)
        with np.printoptions(linewidth=1024, edgeitems=500):
            sol_interp = lbi.eval(t_interp)
        axes[i].plot(t_interp, sol_interp, 'k-')



axes[-1].set_xlabel('time (s)')
# fig.legend((sol[0], sim[0]), ('solution', 'simulation'), 'lower right', ncol=2)
plt.tight_layout()
plt.show()



import numpy as np

import openmdao.api as om

from ..grid_data import GridData
from ...utils.lgl import lgl
from ...utils.lagrange import lagrange_matrices
from ...utils.misc import get_rate_units, reshape_val
from ...utils.constants import INF_BOUND

from ...options import options as dymos_options


class LGLPolynomialControlComp(om.ExplicitComponent):
    """
    Component which interpolates controls as a single polynomial across the entire phase.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('time_units', default=None, allow_none=True, types=str,
                             desc='Units of time')
        self.options.declare('grid_data', types=GridData, desc='Container object for grid info')
        self.options.declare('polynomial_control_options', types=dict,
                             desc='Dictionary of options for the polynomial controls')

        self._matrices = {}

        self._no_check_partials = not dymos_options['include_check_partials']

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine shape and units for the states.
        """
        self._input_names = {}
        self._output_val_names = {}
        self._output_rate_names = {}
        self._output_rate2_names = {}
        self.val_jacs = {}
        self.rate_jacs = {}
        self.rate2_jacs = {}
        self.val_jac_rows = {}
        self.val_jac_cols = {}
        self.rate_jac_rows = {}
        self.rate_jac_cols = {}
        self.rate2_jac_rows = {}
        self.rate2_jac_cols = {}
        self.sizes = {}

        self.add_input('t_duration', val=1.0, units=self.options['time_units'],
                       desc='duration of the phase to which this interpolated control group '
                            'belongs')

        gd = self.options['grid_data']
        num_nodes = gd.subset_num_nodes['all']
        eval_nodes = gd.node_ptau

        for name, options in self.options['polynomial_control_options'].items():
            disc_nodes, _ = lgl(options['order'] + 1)
            num_control_input_nodes = len(disc_nodes)
            shape = options['shape']
            size = np.prod(shape)
            units = options['units']
            rate_units = get_rate_units(units, self.options['time_units'], deriv=1)
            rate2_units = get_rate_units(units, self.options['time_units'], deriv=2)

            input_shape = (num_control_input_nodes,) + shape
            output_shape = (num_nodes,) + shape

            L_de, D_de = lagrange_matrices(disc_nodes, eval_nodes)
            _, D_dd = lagrange_matrices(disc_nodes, disc_nodes)
            D2_de = np.dot(D_de, D_dd)

            self._matrices[name] = L_de, D_de, D2_de

            self._input_names[name] = 'polynomial_controls:{0}'.format(name)
            self._output_val_names[name] = 'polynomial_control_values:{0}'.format(name)
            self._output_rate_names[name] = 'polynomial_control_rates:{0}_rate'.format(name)
            self._output_rate2_names[name] = 'polynomial_control_rates:{0}_rate2'.format(name)

            self.add_input(self._input_names[name], val=np.ones(input_shape), units=units)
            self.add_output(self._output_val_names[name], shape=output_shape, units=units)
            self.add_output(self._output_rate_names[name], shape=output_shape, units=rate_units)
            self.add_output(self._output_rate2_names[name], shape=output_shape, units=rate2_units)

            self.val_jacs[name] = np.zeros((num_nodes, size, num_control_input_nodes, size))
            self.rate_jacs[name] = np.zeros((num_nodes, size, num_control_input_nodes, size))
            self.rate2_jacs[name] = np.zeros((num_nodes, size, num_control_input_nodes, size))

            for i in range(size):
                self.val_jacs[name][:, i, :, i] = L_de
                self.rate_jacs[name][:, i, :, i] = D_de
                self.rate2_jacs[name][:, i, :, i] = D2_de

            self.val_jacs[name] = self.val_jacs[name].reshape((num_nodes * size,
                                                              num_control_input_nodes * size),
                                                              order='C')
            self.rate_jacs[name] = self.rate_jacs[name].reshape((num_nodes * size,
                                                                num_control_input_nodes * size),
                                                                order='C')
            self.rate2_jacs[name] = self.rate2_jacs[name].reshape((num_nodes * size,
                                                                  num_control_input_nodes * size),
                                                                  order='C')
            self.val_jac_rows[name], self.val_jac_cols[name] = \
                np.where(self.val_jacs[name] != 0)
            self.rate_jac_rows[name], self.rate_jac_cols[name] = \
                np.where(self.rate_jacs[name] != 0)
            self.rate2_jac_rows[name], self.rate2_jac_cols[name] = \
                np.where(self.rate2_jacs[name] != 0)

            self.sizes[name] = size

            rs, cs = self.val_jac_rows[name], self.val_jac_cols[name]
            self.declare_partials(of=self._output_val_names[name],
                                  wrt=self._input_names[name],
                                  rows=rs, cols=cs, val=self.val_jacs[name][rs, cs])

            rs = np.concatenate([np.arange(0, num_nodes * size, size, dtype=int) + i
                                 for i in range(size)])

            self.declare_partials(of=self._output_rate_names[name],
                                  wrt='t_duration', rows=rs, cols=np.zeros_like(rs))

            self.declare_partials(of=self._output_rate_names[name],
                                  wrt=self._input_names[name],
                                  rows=self.rate_jac_rows[name], cols=self.rate_jac_cols[name])

            self.declare_partials(of=self._output_rate2_names[name],
                                  wrt='t_duration', rows=rs, cols=np.zeros_like(rs))

            self.declare_partials(of=self._output_rate2_names[name],
                                  wrt=self._input_names[name],
                                  rows=self.rate2_jac_rows[name], cols=self.rate2_jac_cols[name])

    def compute(self, inputs, outputs):
        """
        Interpolate control outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        dt_dptau = 0.5 * inputs['t_duration']

        for name, options in self.options['polynomial_control_options'].items():
            L_de, D_de, D2_de = self._matrices[name]

            u = inputs[self._input_names[name]]

            a = np.tensordot(D_de, u, axes=(1, 0)).T
            b = np.tensordot(D2_de, u, axes=(1, 0)).T

            # divide each "row" by dt_dptau or dt_dptau**2
            outputs[self._output_val_names[name]] = np.tensordot(L_de, u, axes=(1, 0))
            outputs[self._output_rate_names[name]] = (a / dt_dptau).T
            outputs[self._output_rate2_names[name]] = (b / dt_dptau ** 2).T

    def compute_partials(self, inputs, partials):
        """
        Compute sub-jacobian parts. The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        """
        nn = self.options['grid_data'].num_nodes

        for name, options in self.options['polynomial_control_options'].items():
            control_name = self._input_names[name]
            num_input_nodes = options['order'] + 1
            L_de, D_de, D2_de = self._matrices[name]

            size = self.sizes[name]
            rate_name = self._output_rate_names[name]
            rate2_name = self._output_rate2_names[name]

            # Unroll matrix-shaped controls into an array at each node
            u_d = np.reshape(inputs[control_name], (num_input_nodes, size))

            t_duration = inputs['t_duration']
            t_duration_tile = np.tile(t_duration, size * nn)

            partials[rate_name, 't_duration'] = \
                0.5 * (-np.dot(D_de, u_d).ravel(order='F') / (0.5 * t_duration_tile) ** 2)

            partials[rate2_name, 't_duration'] = \
                -1.0 * (np.dot(D2_de, u_d).ravel(order='F') / (0.5 * t_duration_tile) ** 3)

            t_duration_x_size = np.repeat(t_duration, size * nn)[:, np.newaxis]

            r_nz, c_nz = self.rate_jac_rows[name], self.rate_jac_cols[name]
            partials[rate_name, control_name] = \
                (self.rate_jacs[name] / (0.5 * t_duration_x_size))[r_nz, c_nz]

            r_nz, c_nz = self.rate2_jac_rows[name], self.rate2_jac_cols[name]
            partials[rate2_name, control_name] = \
                (self.rate2_jacs[name] / (0.5 * t_duration_x_size) ** 2)[r_nz, c_nz]


class PolynomialControlGroup(om.Group):
    """
    Group that contains and manages the LGLPolynomialControlComp.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def initialize(self):
        """
        Declare group options.
        """
        self.options.declare('polynomial_control_options', types=dict,
                             desc='Dictionary of options for the polynomial controls')
        self.options.declare('time_units', default=None, allow_none=True, types=str,
                             desc='Units of time')
        self.options.declare('grid_data', types=GridData, desc='Container object for grid info')

    def setup(self):
        """
        Define the structure of the polynomial control group.
        """
        opts = self.options

        # Pull out the interpolated controls
        num_opt = 0
        for name, options in opts['polynomial_control_options'].items():
            if options['order'] < 1:
                raise ValueError('Interpolation order must be >= 1 (linear)')
            if options['opt']:
                num_opt += 1

        if num_opt > 0:
            self.add_subsystem('indep_polynomial_controls', subsys=om.IndepVarComp(),
                               promotes_outputs=['*'])

        self.add_subsystem(
            'interp_comp',
            subsys=LGLPolynomialControlComp(time_units=opts['time_units'],
                                            grid_data=opts['grid_data'],
                                            polynomial_control_options=opts['polynomial_control_'
                                                                            'options']),
            promotes_inputs=['*'],
            promotes_outputs=['*'])

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine shape and units for the states.
        """
        self.interp_comp.configure_io()

        # For any interpolated control with `opt=True`, add an indep var comp output and
        # setup the design variable for optimization.
        for name, options in self.options['polynomial_control_options'].items():
            num_input_nodes = options['order'] + 1
            shape = options['shape']
            if options['opt']:
                default_val = reshape_val(options['val'], shape, num_input_nodes)

                self.indep_polynomial_controls.add_output(f'polynomial_controls:{name}',
                                                          shape=(num_input_nodes,) + shape,
                                                          val=default_val,
                                                          units=options['units'])

                desvar_indices = list(range(num_input_nodes))
                if options['fix_initial']:
                    desvar_indices.pop(0)
                if options['fix_final']:
                    desvar_indices.pop()

                lb = -INF_BOUND if options['lower'] is None else options['lower']
                ub = INF_BOUND if options['upper'] is None else options['upper']

                self.add_design_var('polynomial_controls:{0}'.format(name),
                                    lower=lb,
                                    upper=ub,
                                    ref=options['ref'],
                                    ref0=options['ref0'],
                                    adder=options['adder'],
                                    scaler=options['scaler'],
                                    indices=desvar_indices,
                                    flat_indices=True)



import numpy as np
import openmdao.api as om

from ...options import options as dymos_options


class PathConstraintComp(om.ExplicitComponent):
    """
    Component that computes path constraint values.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self._path_constraints = []
        self._vars = []
        self.options.declare('num_nodes', types=(int,), desc='The number of nodes in the phase '
                                                             'at which the path constraint is to '
                                                             'be evaluated')

    def setup(self):
        """
        Define the independent variables as output variables.
        """
        for (name, kwargs) in self._path_constraints:
            self._add_path_constraint_configure(name=name, shape=kwargs['shape'],
                                                units=kwargs['units'], desc=kwargs['desc'],
                                                indices=kwargs['indices'], lower=kwargs['lower'],
                                                upper=kwargs['upper'], equals=kwargs['equals'],
                                                scaler=kwargs['scaler'], adder=kwargs['adder'],
                                                ref0=kwargs['ref0'], ref=kwargs['ref'],
                                                linear=kwargs['linear'])

    def _add_path_constraint_configure(self, name, shape=None, units=None,
                                       desc='', indices=None, lower=None, upper=None, equals=None,
                                       scaler=None, adder=None, ref=None, ref0=None, linear=False):
        """
        Add a path constraint to this component during the configure portion of the setup stack.

        Parameters
        ----------
        name : str
            name of the variable in this component's namespace.
        shape : int or tuple or list or None
            Shape of this variable, only required if val is not an array.
            Default is None.
        indices : tuple or list or ndarray or None
            Indices that specify which elements in shape are to be path constrained.  If None,
            then the constraint will apply to all values and lower/upper/equals must be scalar
            or of the same shape.  Indices assumes C-order flattening.  For instance, if
            constraining element [0, 1] of a variable with shape [2, 2], indices=[3].
        units : str or None
            Units in which the output variables will be provided to the component during execution.
            Default is None, which means it has no units.
        desc : str
            description of the variable
        lower : float or list or tuple or ndarray or None
            lower bound(s) in user-defined units. It can be (1) a float, (2) an array_like
            consistent with the shape arg (if given), or (3) an array_like matching the shape of
            val, if val is array_like. A value of None means this output has no lower bound.
            Default is None.
        upper : float or list or tuple or ndarray or None
            upper bound(s) in user-defined units. It can be (1) a float, (2) an array_like
            consistent with the shape arg (if given), or (3) an array_like matching the shape of
            val, if val is array_like. A value of None means this output has no upper bound.
            Default is None.
        ref : float
            Scaling parameter. The value in the user-defined units of this output variable when
            the scaled value is 1. Default is 1.
        ref0 : float
            Scaling parameter. The value in the user-defined units of this output variable when
            the scaled value is 0. Default is 0.
        linear : bool
            True if the *total* derivative of the constrained variable is linear, otherwise False.
        """
        num_nodes = self.options['num_nodes']

        input_name = f'all_values:{name}'
        self.add_input(input_name, shape=(num_nodes,) + shape, units=units, desc=desc)

        output_name = f'path:{name}'
        self.add_output(output_name, shape=(num_nodes,) + shape, units=units, desc=desc)

        # Convert indices from those in one time instance to those in all time instances
        template = np.zeros(np.prod(shape), dtype=int)
        template[indices] = 1
        template = np.tile(template, num_nodes)
        indices = np.nonzero(template)[0]

        self.add_constraint(output_name, lower=lower, upper=upper, equals=equals, ref0=ref0,
                            ref=ref, scaler=scaler, adder=adder, indices=indices, flat_indices=True,
                            linear=linear)

        self._vars.append((input_name, output_name, shape))

        # Setup partials
        all_shape = (num_nodes,) + shape
        var_size = np.prod(shape)
        all_size = np.prod(all_shape)

        all_row_starts = np.arange(num_nodes, dtype=int) * var_size
        all_rows = []
        for i in all_row_starts:
            all_rows.extend(range(i, i + var_size))
        all_rows = np.asarray(all_rows, dtype=int)

        self.declare_partials(of=output_name, wrt=input_name, rows=all_rows,
                              cols=np.arange(all_size), val=1.0)

    def compute(self, inputs, outputs):
        """
        Compute path constraint outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        for (input_name, output_name, _) in self._vars:
            outputs[output_name] = inputs[input_name]



import functools

import numpy as np

from scipy.interpolate import interp1d

import openmdao.utils.assert_utils as _om_assert_utils


def assert_check_partials(data, atol=1.0E-6, rtol=1.0E-6):
    """
    Wrapper around OpenMDAO's assert_check_partials with a dymos-specific message.

    Calls OpenMDAO's assert_check_partials but verifies that the dictionary of assertion data is
    not empty due to dymos.options['include_check_partials'] being False.

    Parameters
    ----------
    data : dict of dicts of dicts
            First key:
                is the component name;
            Second key:
                is the (output, input) tuple of strings;
            Third key:
                is one of ['rel error', 'abs error', 'magnitude', 'J_fd', 'J_fwd', 'J_rev'];

            For 'rel error', 'abs error', 'magnitude' the value is: A tuple containing norms for
                forward - fd, adjoint - fd, forward - adjoint.
            For 'J_fd', 'J_fwd', 'J_rev' the value is: A numpy array representing the computed
                Jacobian for the three different methods of computation.
    atol : float
        Absolute error. Default is 1e-6.
    rtol : float
        Relative error. Default is 1e-6.
    """
    assert len(data) >= 1, "No check partials data found.  Is " \
                           "dymos.options['include_check_partials'] set to True?"
    _om_assert_utils.assert_check_partials(data, atol, rtol)


def assert_cases_equal(case1, case2, tol=1.0E-12, require_same_vars=True):
    """
    Raise AssertionError if the data in two OpenMDAO Cases is different.

    Parameters
    ----------
    case1 : om.Case
        The first OpenMDAO Case for comparison.
    case2 : om.Case
        The second OpenMDAO Case for comparison.
    tol : float
        The absolute value of the allowable difference in values between two variables.
    require_same_vars : bool
        If True, require that the two files contain the same set of variables.

    Raises
    ------
    AssertionError
        Raised in the following cases:  If require_same_vars is True, then AssertionError is raised
        if the two cases contain different variables.  Otherwise, this error is raised if case1
        and case2 contain the same variable but the variable has a different size/shape in the two
        cases, or if the variables have the same shape but different values (as given by tol).
    """
    case1_vars = {t[1]['prom_name']: t[1] for t in
                  case1.list_inputs(values=True, units=True, prom_name=True, out_stream=None)}
    case1_vars.update({t[1]['prom_name']: t[1] for t in
                       case1.list_outputs(values=True, units=True, prom_name=True, out_stream=None)})

    case2_vars = {t[1]['prom_name']: t[1] for t in
                  case2.list_inputs(values=True, units=True, prom_name=True, out_stream=None)}
    case2_vars.update({t[1]['prom_name']: t[1] for t in
                       case2.list_outputs(values=True, units=True, prom_name=True, out_stream=None)})

    # Warn if a and b don't contain the same sets of variables
    diff_err_msg = ''
    if require_same_vars:
        case1_minus_case2 = set(case1_vars.keys()) - set(case2_vars.keys())
        case2_minus_case1 = set(case2_vars.keys()) - set(case1_vars.keys())
        if case1_minus_case2 or case2_minus_case1:
            diff_err_msg = '\nrequire_same_vars=True but cases contain different variables.'
        if case1_minus_case2:
            diff_err_msg += f'\nVariables in case1 but not in case2: {sorted(case1_minus_case2)}'
        if case2_minus_case1:
            diff_err_msg += f'\nVariables in case2 but not in case1: {sorted(case2_minus_case1)}'

    shape_errors = set()
    val_errors = set()
    shape_err_msg = '\nThe following variables have different shapes/sizes:'
    val_err_msg = '\nThe following variables contain different values:\nvar: error'

    for var in sorted(set(case1_vars.keys()).intersection(case2_vars.keys())):
        a = case1_vars[var]['value']
        b = case2_vars[var]['value']
        if a.shape != b.shape:
            shape_errors.add(var)
            shape_err_msg += f'\n{var} has shape {a.shape} in case1 but shape {b.shape} in case2'
            continue
        err = np.abs(a - b)
        if np.any(err > tol):
            val_errors.add(var)
            val_err_msg += f'\n{var}: {err}'

    err_msg = ''
    if diff_err_msg:
        err_msg += diff_err_msg
    if shape_errors:
        err_msg += shape_err_msg
    if val_errors:
        err_msg += val_err_msg

    if err_msg:
        raise AssertionError(err_msg)


def assert_timeseries_near_equal(t1, x1, t2, x2, tolerance=1.0E-6):
    """
    Assert that two timeseries of data are approximately equal.

    This is done by fitting a 1D interpolant to each index of each timeseries, and then comparing
    the values of the two interpolants at some equally spaced number of points.

    Parameters
    ----------
    t1 : np.array
        Time values for the first timeseries.
    x1 : np.array
        Data values for the first timeseries.
    t2 : np.array
        Time values for the second timeseries.
    x2 : np.array
        Data values for the second timeseries.
    tolerance : float
        The tolerance for any errors along at each point checked.

    Raises
    ------
    AssertionError
        When one or more elements of the interpolated timeseries are not within the
        desired tolerance.
    """
    shape1 = x1.shape[1:]
    shape2 = x2.shape[1:]

    if shape1 != shape2:
        raise ValueError('The shape of the variable in the two timeseries is not equal '
                         f'x1 is {shape1}  x2 is {shape2}')

    if abs(t1[0] - t2[0]) > 1.0E-12:
        raise ValueError('The initial time of the two timeseries is not the same. '
                         f't1[0]={t1[0]}  t2[0]={t2[0]}  difference: {t2[0] - t1[0]}')

    if abs(t1[-1] - t2[-1]) > 1.0E-12:
        raise ValueError('The final time of the two timeseries is not the same. '
                         f't1[0]={t1[-1]}  t2[0]={t2[-1]}  difference: {t2[-1] - t1[-1]}')

    size = np.prod(shape1)

    nn1 = x1.shape[0]
    a1 = np.reshape(x1, newshape=(nn1, size))
    t1_unique, idxs1 = np.unique(t1.ravel(), return_index=True)

    nn2 = x2.shape[0]
    a2 = np.reshape(x2, newshape=(nn2, size))
    t2_unique, idxs2 = np.unique(t2.ravel(), return_index=True)

    if nn1 > nn2:
        # The first timeseries is more dense
        t_unique = t1_unique
        x_to_interp = a1[idxs1, ...]
        t_check = t2.ravel()
        x_check = x2
    else:
        # The second timeseries is more dense
        t_unique = t2_unique
        x_to_interp = a2[idxs2, ...]
        t_check = t1.ravel()
        x_check = x1

    interp = interp1d(x=t_unique, y=x_to_interp, kind='slinear', axis=0)
    num_points = np.prod(t_check.shape)

    y_interp = np.reshape(interp(t_check), newshape=(num_points,) + shape1)

    _om_assert_utils.assert_near_equal(y_interp, x_check, tolerance=tolerance)



"""
Integration test for a battery+motor example that demonstrates phase branching in trajectories.
"""
import unittest

import matplotlib
matplotlib.use('Agg')

from openmdao.utils.testing_utils import use_tempdirs


@use_tempdirs
class TestBatteryBranchingPhasesForDocs(unittest.TestCase):

    def test_basic(self):
        import matplotlib.pyplot as plt

        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal

        import dymos as dm
        from dymos.examples.battery_multibranch.battery_multibranch_ode import BatteryODE
        from dymos.utils.lgl import lgl

        prob = om.Problem()

        opt = prob.driver = om.ScipyOptimizeDriver()
        opt.declare_coloring()
        opt.options['optimizer'] = 'SLSQP'

        num_seg = 5
        seg_ends, _ = lgl(num_seg + 1)

        traj = prob.model.add_subsystem('traj', dm.Trajectory())

        # First phase: normal operation.
        transcription = dm.Radau(num_segments=num_seg, order=5, segment_ends=seg_ends, compressed=False)
        phase0 = dm.Phase(ode_class=BatteryODE, transcription=transcription)
        traj_p0 = traj.add_phase('phase0', phase0)

        traj_p0.set_time_options(fix_initial=True, fix_duration=True)
        traj_p0.add_state('state_of_charge', fix_initial=True, fix_final=False,
                          targets=['SOC'], rate_source='dXdt:SOC')

        # Second phase: normal operation.

        phase1 = dm.Phase(ode_class=BatteryODE, transcription=transcription)
        traj_p1 = traj.add_phase('phase1', phase1)

        traj_p1.set_time_options(fix_initial=False, fix_duration=True)
        traj_p1.add_state('state_of_charge', fix_initial=False, fix_final=False,
                          targets=['SOC'], rate_source='dXdt:SOC')
        traj_p1.add_objective('time', loc='final')

        # Second phase, but with battery failure.

        phase1_bfail = dm.Phase(ode_class=BatteryODE, ode_init_kwargs={'num_battery': 2},
                                transcription=transcription)
        traj_p1_bfail = traj.add_phase('phase1_bfail', phase1_bfail)

        traj_p1_bfail.set_time_options(fix_initial=False, fix_duration=True)
        traj_p1_bfail.add_state('state_of_charge', fix_initial=False, fix_final=False,
                                targets=['SOC'], rate_source='dXdt:SOC')

        # Second phase, but with motor failure.

        phase1_mfail = dm.Phase(ode_class=BatteryODE, ode_init_kwargs={'num_motor': 2},
                                transcription=transcription)
        traj_p1_mfail = traj.add_phase('phase1_mfail', phase1_mfail)

        traj_p1_mfail.set_time_options(fix_initial=False, fix_duration=True)
        traj_p1_mfail.add_state('state_of_charge', fix_initial=False, fix_final=False,
                                targets=['SOC'], rate_source='dXdt:SOC')

        traj.link_phases(phases=['phase0', 'phase1'], vars=['state_of_charge', 'time'])
        traj.link_phases(phases=['phase0', 'phase1_bfail'], vars=['state_of_charge', 'time'])
        traj.link_phases(phases=['phase0', 'phase1_mfail'], vars=['state_of_charge', 'time'])

        prob.model.options['assembled_jac_type'] = 'csc'
        prob.model.linear_solver = om.DirectSolver(assemble_jac=True)

        prob.setup()

        prob['traj.phase0.t_initial'] = 0
        prob['traj.phase0.t_duration'] = 1.0*3600

        prob['traj.phase1.t_initial'] = 1.0*3600
        prob['traj.phase1.t_duration'] = 1.0*3600

        prob['traj.phase1_bfail.t_initial'] = 1.0*3600
        prob['traj.phase1_bfail.t_duration'] = 1.0*3600

        prob['traj.phase1_mfail.t_initial'] = 1.0*3600
        prob['traj.phase1_mfail.t_duration'] = 1.0*3600

        prob.set_solver_print(level=0)
        dm.run_problem(prob)

        soc0 = prob['traj.phase0.states:state_of_charge']
        soc1 = prob['traj.phase1.states:state_of_charge']
        soc1b = prob['traj.phase1_bfail.states:state_of_charge']
        soc1m = prob['traj.phase1_mfail.states:state_of_charge']

        # Final value for State of Chrage in each segment should be a good test.
        print('State of Charge after 1 hour')
        assert_near_equal(soc0[-1], 0.63464982, 1e-6)
        print('State of Charge after 2 hours')
        assert_near_equal(soc1[-1], 0.23794217, 1e-6)
        print('State of Charge after 2 hours, battery fails at 1 hour')
        assert_near_equal(soc1b[-1], 0.0281523, 1e-6)
        print('State of Charge after 2 hours, motor fails at 1 hour')
        assert_near_equal(soc1m[-1], 0.18625395, 1e-6)

        # Plot Results
        t0 = prob['traj.phases.phase0.time.time']/3600
        t1 = prob['traj.phases.phase1.time.time']/3600
        t1b = prob['traj.phases.phase1_bfail.time.time']/3600
        t1m = prob['traj.phases.phase1_mfail.time.time']/3600

        plt.subplot(2, 1, 1)
        plt.plot(t0, soc0, 'b')
        plt.plot(t1, soc1, 'b')
        plt.plot(t1b, soc1b, 'r')
        plt.plot(t1m, soc1m, 'c')
        plt.xlabel('Time (hour)')
        plt.ylabel('State of Charge (percent)')

        I_Li0 = prob['traj.phases.phase0.rhs_all.pwr_balance.I_Li']
        I_Li1 = prob['traj.phases.phase1.rhs_all.pwr_balance.I_Li']
        I_Li1b = prob['traj.phases.phase1_bfail.rhs_all.pwr_balance.I_Li']
        I_Li1m = prob['traj.phases.phase1_mfail.rhs_all.pwr_balance.I_Li']

        plt.subplot(2, 1, 2)
        plt.plot(t0, I_Li0, 'b')
        plt.plot(t1, I_Li1, 'b')
        plt.plot(t1b, I_Li1b, 'r')
        plt.plot(t1m, I_Li1m, 'c')
        plt.xlabel('Time (hour)')
        plt.ylabel('Line Current (A)')

        plt.legend(['Phase 1', 'Phase 2', 'Phase 2 Battery Fail', 'Phase 2 Motor Fail'], loc=2)

        plt.show()


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import warnings

import openmdao.api as om
from dymos.trajectory.trajectory import Trajectory
from dymos.load_case import load_case
from dymos.visualization.timeseries_plots import timeseries_plots

from .grid_refinement.refinement import _refine_iter


def run_problem(problem, refine_method='hp', refine_iteration_limit=0, run_driver=True,
                simulate=False, restart=None,
                solution_record_file='dymos_solution.db',
                simulation_record_file='dymos_simulation.db',
                make_plots=False,
                plot_dir="plots"
                ):
    """
    A Dymos-specific interface to execute an OpenMDAO problem containing Dymos Trajectories or
    Phases.  This function can iteratively call run_driver to perform grid refinement, and
    automatically call simulate following a run to check the validity of a result.

    Parameters
    ----------
    problem : om.Problem
        The OpenMDAO problem object to be run.
    refine_method : String
        The choice of refinement algorithm to use for grid refinement
    refine_iteration_limit : int
        The number of passes through the grid refinement algorithm to be made.
    run_driver : bool
        If True, run the driver (optimize the problem), otherwise just run the model one time.
    simulate : bool
        If True, perform a simulation of Trajectories found in the Problem after the driver
        has been run and grid refinement is complete.
    restart : str
        If not None, the path to a CaseRecorder file used to load in recording from a previous run.
    make_plots : bool
        If True, automatically generate plots of all timeseries outputs.
    solution_record_file : String
        Path to case recorder file use to store results from solution.
    simulation_record_file : String
        Path to case recorder file use to store results from simulation.
    plot_dir : str
        Path to directory for plot files.
    """
    if restart is not None:
        case = om.CaseReader(restart).get_case('final')

    if solution_record_file not in [rec._filepath for rec in iter(problem._rec_mgr)]:
        recorder = om.SqliteRecorder(solution_record_file)
        problem.add_recorder(recorder)
        # record_inputs is needed to capture potential input parameters that aren't connected
        problem.recording_options['record_inputs'] = True
        # record_outputs is need to capture the timeseries outputs
        problem.recording_options['record_outputs'] = True

    problem.final_setup()

    if restart is not None:
        load_case(problem, case)

    if run_driver:
        failed = problem.run_driver()
        _refine_iter(problem, refine_iteration_limit, refine_method)
    else:
        failed = problem.run_model()
        if refine_iteration_limit > 0:
            warnings.warn("Refinement not performed. Set run_driver to True to perform refinement.")

    problem.record('final')  # save case for potential restart
    problem.cleanup()

    if simulate:
        for subsys in problem.model.system_iter(include_self=True, recurse=True):
            if isinstance(subsys, Trajectory):
                subsys.simulate(record_file=simulation_record_file)

    if make_plots:
        if simulate:
            timeseries_plots(solution_record_file, simulation_record_file=simulation_record_file,
                             plot_dir=plot_dir)
        else:
            timeseries_plots(solution_record_file, simulation_record_file=None,
                             plot_dir=plot_dir)

    return failed



import os
import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs
import dymos as dm

from dymos.utils.testing_utils import assert_cases_equal, assert_timeseries_near_equal


@use_tempdirs
class TestAssertCasesEqual(unittest.TestCase):

    def tearDown(self):
        for file in ('p1.db', 'p2.db'):
            try:
                os.remove(file)
                print('removed', file)
            except:
                print(f'no file named {file}')

    def test_different_variables(self):

        p1 = om.Problem()
        ivc = p1.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=0.0)
        ivc.add_output('b', val=1.0)
        ivc.add_output('c', val=2.0)
        ivc.add_output('x', val=3.0)
        p1.add_recorder(om.SqliteRecorder('p1.db'))
        p1.setup()

        p2 = om.Problem()
        ivc = p2.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('x', val=3.0)
        ivc.add_output('y', val=4.0)
        ivc.add_output('z', val=5.0)
        p2.add_recorder(om.SqliteRecorder('p2.db'))
        p2.setup()

        p1.run_model()
        p2.run_model()

        p1.record('final')
        p1.cleanup()

        p2.record('final')
        p2.cleanup()

        c1 = om.CaseReader('p1.db').get_case('final')
        c2 = om.CaseReader('p2.db').get_case('final')

        with self.assertRaises(AssertionError) as e:
            assert_cases_equal(c1, c2)

        expected = "\nrequire_same_vars=True but cases contain different variables.\nVariables in " \
                   "case1 but not in case2: ['a', 'b', 'c']\nVariables in case2 but not in " \
                   "case1: ['y', 'z']"

        self.assertEqual(str(e.exception), expected)

    def test_allow_different_variables(self):
        p1 = om.Problem()
        ivc = p1.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=0.0)
        ivc.add_output('b', val=1.0)
        ivc.add_output('c', val=2.0)
        ivc.add_output('x', val=3.0)
        p1.add_recorder(om.SqliteRecorder('p1.db'))
        p1.setup()

        p2 = om.Problem()
        ivc = p2.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('x', val=3.0)
        ivc.add_output('y', val=4.0)
        ivc.add_output('z', val=5.0)
        p2.add_recorder(om.SqliteRecorder('p2.db'))
        p2.setup()

        p1.run_model()
        p2.run_model()

        p1.record('final')
        p1.cleanup()

        p2.record('final')
        p2.cleanup()

        c1 = om.CaseReader('p1.db').get_case('final')
        c2 = om.CaseReader('p2.db').get_case('final')

        assert_cases_equal(c1, c2, require_same_vars=False)

    def test_different_shapes(self):

        p1 = om.Problem()
        ivc = p1.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=np.array([0, 1, 2]))
        ivc.add_output('b', val=np.array([[0, 1, 2], [3, 4, 5]]))
        ivc.add_output('c', val=np.eye(3, dtype=float))
        p1.add_recorder(om.SqliteRecorder('p1.db'))
        p1.setup()

        p2 = om.Problem()
        ivc = p2.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=3.0)
        ivc.add_output('b', val=np.array([0, 1, 2]))
        ivc.add_output('c', val=np.ones(3, dtype=float))
        p2.add_recorder(om.SqliteRecorder('p2.db'))
        p2.setup()

        p1.run_model()
        p2.run_model()

        p1.record('final')
        p1.cleanup()

        p2.record('final')
        p2.cleanup()

        c1 = om.CaseReader('p1.db').get_case('final')
        c2 = om.CaseReader('p2.db').get_case('final')

        with self.assertRaises(AssertionError) as e:
            assert_cases_equal(c1, c2)

        expected = "\nThe following variables have different shapes/sizes:\na has shape (3,) in " \
                   "case1 but shape (1,) in case2\nb has shape (2, 3) in case1 but shape (3,) in " \
                   "case2\nc has shape (3, 3) in case1 but shape (3,) in case2"

        self.assertEqual(str(e.exception), expected)

    def test_different_values(self):

        p1 = om.Problem()
        ivc = p1.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=np.array([0, 1, 2]))
        ivc.add_output('b', val=np.array([[0, 1, 2], [3, 4, 5]]))
        ivc.add_output('c', val=np.eye(3, dtype=float))
        p1.add_recorder(om.SqliteRecorder('p1.db'))
        p1.setup()

        p2 = om.Problem()
        ivc = p2.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('a', val=3 * np.array([0, 1, 2]))
        ivc.add_output('b', val=2 * np.array([[0, 1, 2], [3, 4, 5]]))
        ivc.add_output('c', val=5 * np.eye(3, dtype=float))
        p2.add_recorder(om.SqliteRecorder('p2.db'))
        p2.setup()

        p1.run_model()
        p2.run_model()

        p1.record('final')
        p1.cleanup()

        p2.record('final')
        p2.cleanup()

        c1 = om.CaseReader('p1.db').get_case('final')
        c2 = om.CaseReader('p2.db').get_case('final')

        expected = "\nThe following variables contain different values:\nvar: " \
                   "error\na: [0. 2. 4.]\nb: [[0. 1. 2.]\n [3. 4. 5.]]\n" \
                   "c: [[4. 0. 0.]\n [0. 4. 0.]\n [0. 0. 4.]]"

        with self.assertRaises(AssertionError) as e:
            assert_cases_equal(c1, c2)

        self.assertEqual(str(e.exception), expected)


@use_tempdirs
class TestAssertTimeseriesNearEqual(unittest.TestCase):

    def test_assert_different_shape(self):

        t1 = np.linspace(0, 100, 50)
        t2 = np.linspace(0, 100, 60)

        x1 = np.atleast_2d(np.sin(t1)).T
        x2 = np.atleast_3d(np.sin(t2)).T

        with self.assertRaises(ValueError) as e:
            assert_timeseries_near_equal(t1, x1, t2, x2)

        expected = "The shape of the variable in the two timeseries is not equal x1 is (1,)  x2 is (60, 1)"

        self.assertEqual(expected, str(e.exception))

    def test_assert_different_initial_time(self):

        t1 = np.linspace(0, 100, 50)
        t2 = np.linspace(5, 100, 50)

        x1 = np.atleast_2d(np.sin(t1)).T
        x2 = np.atleast_2d(np.sin(t2)).T

        with self.assertRaises(ValueError) as e:
            assert_timeseries_near_equal(t1, x1, t2, x2)

        expected = "The initial time of the two timeseries is not the same. t1[0]=0.0  " \
                   "t2[0]=5.0  difference: 5.0"

        self.assertEqual(str(e.exception), expected)

    def test_assert_different_final_time(self):

        t1 = np.linspace(0, 100, 50)
        t2 = np.linspace(0, 102, 50)

        x1 = np.atleast_2d(np.sin(t1)).T
        x2 = np.atleast_2d(np.sin(t2)).T

        with self.assertRaises(ValueError) as e:
            assert_timeseries_near_equal(t1, x1, t2, x2)

        expected = "The final time of the two timeseries is not the same. t1[0]=100.0  " \
                   "t2[0]=102.0  difference: 2.0"

        self.assertEqual(str(e.exception), expected)

    def test_assert_different_values(self):

        t1 = np.linspace(0, 100, 50)
        t2 = np.linspace(0, 100, 50)

        x1 = np.atleast_2d(np.sin(t1)).T
        x2 = np.atleast_2d(np.cos(t2)).T

        with self.assertRaises(ValueError) as e:
            assert_timeseries_near_equal(t1, x1, t2, x2)
